/* Stats - IRC Statistical Services for Hybrid, csircd, and IRCnet 2.10
** evh.c - Event handlers
**
** Copyright W. Campbell and others.  See README for more details
** Some code Copyright: Jonathan George, Kai Seidler, ircd-hybrid Team,
**                      IRCnet IRCD developers.
**
** $Id: evh.c,v 1.20 2002/10/23 20:07:53 wcampbel Exp $
*/

#include "stats.h"
#include "struct.h"
#include "proto.h"
#include "event.h"

static void evh_CheckRehash(void *unused);
static void evh_StatsSave(void *unused);
#ifdef DIE_ON_MASS_COLLIDE
static void evh_ResetCollides(void *unused);
#endif
static void evh_Ping(void *unused);
static void evh_HTML(void *unused);
static void evh_DataDump(void *unused);
static void evh_CheckMidnight(void *unused);
#if defined(SPLITSERV)
static void evh_CheckDelayed(void *unused);
#endif
#ifdef SPLITSERV
static void evh_ExpireSplits(void *unused);
static void evh_DumpMap(void *unused);
#endif
static void evh_Delayed(void *unused);
static void evh_FlushLogs(void *unused);

/* This function installs all continuously running events and sets
** up the handler for "delayed" events
*/
void StartInitialEvents(void)
{
	eventAdd("check rehash", evh_CheckRehash, NULL, REHASHINTERVAL);
	eventAdd("stats save", evh_StatsSave, NULL, STATSSAVETIME);
#ifdef DIE_ON_MASS_COLLIDE
	eventAdd("reset collisions", evh_ResetCollides, NULL, COLLIDETIME);
#endif
	eventAdd("uplink ping", evh_Ping, NULL, PINGDELAY);

	/* Once per second (by default) for now, see what the performance
	** hit would be
	*/
	eventAdd("check midnight", evh_CheckMidnight, NULL, MIDNIGHTINTERVAL);
#if defined(SPLITSERV)
	eventAdd("check delayed", evh_CheckDelayed, NULL, sp.delay);
#endif
#if defined(SPYMODE) || defined(ALERTMODE)
	eventAdd("check admin delayed", evh_CheckAdDelayed, NULL,DEFAULT_DELAY);
#endif
#ifdef SPLITSERV
	eventAdd("expire splits", evh_ExpireSplits, NULL, EXPIREPOLL);
#endif
	eventAdd("flush logfiles", evh_FlushLogs, NULL, LOGFLUSHTIME);
	
	eventAdd("delayed events", evh_Delayed, NULL, DEFAULT_DELAY);
}

void FlushEvents(void)
{
	eventDelete(evh_CheckRehash, NULL);
	eventDelete(evh_StatsSave, NULL);
#ifdef DIE_ON_MASS_COLLIDE
	eventDelete(evh_ResetCollides, NULL);
#endif
	eventDelete(evh_Ping, NULL);
#if defined(SPLITSERV)
	eventDelete(evh_CheckDelayed, NULL);
#endif
#if defined(SPYMODE) || defined(ALERTMODE)
	eventDelete(evh_CheckAdDelayed, NULL);
#endif
#ifdef SPLITSERV
	eventDelete(evh_ExpireSplits, NULL);
#endif
	eventDelete(evh_CheckMidnight, NULL);
	eventDelete(evh_FlushLogs, NULL);
	eventDelete(evh_Delayed, NULL);
	eventDelete(evh_HTML, NULL);
	eventDelete(evh_DataDump, NULL);
#ifdef SPLITSERV
	eventDelete(evh_DumpMap, NULL);
#endif
}

void ResetTimedEvents(void)
{
	/* Remove the current calls of these events */
	eventDelete(evh_HTML, NULL);
	eventDelete(evh_DataDump, NULL);
#ifdef SPLITSERV
	eventDelete(evh_DumpMap, NULL);
#endif
	/* Then restart them with the new times */
	evh_Delayed(NULL);

	/* and finally deal with CheckDelayed... */
#ifdef SPLITSERV
	eventDelete(evh_CheckDelayed, NULL);
	eventAdd("check delayed", evh_CheckDelayed, NULL, sp.delay);
#endif

#if defined(SPYMODE) || defined(ALERTMODE)
	eventDelete(evh_CheckAdDelayed, NULL);
	eventAdd("check admin delayed", evh_CheckAdDelayed, NULL,DEFAULT_DELAY);
#endif
}

static void evh_Delayed(void *unused)
{
	evh_HTML(NULL);
	eventAdd("HTML dump", evh_HTML, NULL, settings.html_time);

	evh_DataDump(NULL);
	eventAdd("data dump", evh_DataDump, NULL, settings.dump_time);

#ifdef SPLITSERV
	evh_DumpMap(NULL);
	eventAdd("dump map text", evh_DumpMap, NULL, sp.text_time);
#endif

	/* And remove myself */
	eventDelete(evh_Delayed, NULL);
}

static void evh_CheckRehash(void *unused)
{
	if (need_rehash)
	{
		rehash();
                log("SIGHUP Caught, REHASH completed");
                alert("SIGHUP Caught, REHASH completed");
		sendto_dcc_flag(CF_NOTICES, "SIGHUP Caught, REHASH completed");
                need_rehash = 0;
                wallops(settings.name, "SIGHUP Caught, REHASH completed");
        }
}

static void evh_StatsSave(void *unused)
{
	SaveStats();
}

#ifdef DIE_ON_MASS_COLLIDE
static void evh_ResetCollides(void *unused)
{
	/* netinfo.collides = 0; */
	/* For better "security", I'll simply decrement it */
	if (netinfo.collides > 0)
		netinfo.collides--;
}
#endif

static void evh_Ping(void *unused)
{
	if (me.s != NULL)
		sts("PING %s", me.s->name);
}

static void evh_HTML(void *unused)
{
	if (IsHtml())
	{
		DumpHTML();
		spy("HTML Dump Completed");
		sendto_dcc_flag(CF_MISC, "HTML Dump Completed");
	}
}

static void evh_DataDump(void *unused)
{
	if (IsDump())
	{
		datadump(settings.dump_file);
		spy("Data Dump Completed");
		sendto_dcc_flag(CF_MISC, "Data Dump Completed");
	}
}

static void evh_CheckMidnight(void *unused)
{
	static int midnight = 0;

	if (!midnight && is_midnight())
	{
		/* Perform all rollover tasks */
		time_t current = time(NULL);
		debug("Its midnight!!! -> %s", sctime(current));
		spy("Local Midnight (or close to it) Rollover");
		sendto_dcc_flag(CF_MISC,
			"Local Midnight (or close to it) Rollover");
		/* Reset Daily Statistics */
		daily.servers = netinfo.servers;
		daily.t_servers = current;
		daily.users = netinfo.users;
		daily.t_users = current;
		daily.opers = netinfo.opers;
		daily.t_opers = current;
		/* end */
		ResetLogs();

		/* Then to make sure it doesn't get run again... */
		midnight = 1;
	}
	else
	{
		if (midnight && !(is_midnight()))
			midnight = 0;
	}
}

#if defined(SPLITSERV)
static void evh_CheckDelayed(void *unused)
{
	is_delayed = 0;
	eventDelete(evh_CheckDelayed, NULL);
}
#endif

#if defined(SPYMODE) || defined(ALERTMODE)
void evh_CheckAdDelayed(void *unused)
{
	ad_delayed = 0;
	eventDelete(evh_CheckAdDelayed, NULL);
}
#endif

#ifdef SPLITSERV
static void evh_ExpireSplits(void *unused)
{
	if (IsSPEnableExpire())
		expire_splits();
}

static void evh_DumpMap(void *unused)
{
	if (IsSPEnableText())
		dump_map();
}
#endif

static void evh_FlushLogs(void *unused)
{
	if (main_log != NULL)
	{
		fflush(main_log);
		fclose(main_log);
		main_log = fopen(LOGFILE, "a");
		/* It's ok if main_log stays NULL */
	}
#ifdef SPLITSERV
	if (split_log != NULL)
	{
		fflush(split_log);
		fclose(split_log);
		split_log = fopen(SPLITLOG, "a");
	}
#endif
#ifdef IO_LOG
	if (io_log != NULL)
	{
		fflush(io_log);
		fclose(io_log);
		io_log = fopen(IOLOGFILE, "a");
	}
#endif
}
