/* Stats - IRC Statistical Services for Hybrid, csircd, and IRCnet 2.10
** ircd.c - ircd parsing related functions
**
** Copyright W. Campbell and others.  See README for more details
** Some code Copyright: Jonathan George, Kai Seidler, ircd-hybrid Team,
**                      IRCnet IRCD developers.
**
** $Id: ircd.c,v 1.193 2002/10/31 13:54:36 wcampbel Exp $
*/

#include "stats.h"
#include "struct.h"
#include "proto.h"
#include "numeric.h"
#include "text.h"
#include "event.h"

extern const char *infotext[];
static char errstr[BUFSIZE];

static void ShowMOTD(char *);
static void HandleCollide();
static void CollideDie();
static void Trace(struct User *);
static void ParseStats(struct User *, char);

#ifdef SUPPORT_MSG_OPERS
static void MsgOpers(char *, char *);
#endif

static int m_time(char *, int, char **);
static int m_version(char *, int, char **);
static int m_motd(char *, int, char **);
static int m_privmsg(char *, int, char **);
static int m_notice(char *, int, char **);
static int m_server(char *, int, char **);
static int m_squit(char *, int, char **);
static int m_quit(char *, int, char **);
#ifdef TRACK_VERSION
static int m_351(char *, int, char **);
#endif
#ifdef TRACK_ADMIN
static int m_257(char *, int, char **);
static int m_258(char *, int, char **);
static int m_259(char *, int, char **);
#endif
static int m_mode(char *, int, char **);
static int m_kill(char *, int, char **);
#if defined(IRCD_HYB7) || defined(IRCD_BAHAMUT) || defined(IRCD_HYBSVS)
static int m_away(char *, int, char **);
#endif
static int m_nick(char *, int, char **);
#ifdef IRCD_2_10
static int m_service(char *, int, char **);
#endif
static int m_ping(char *, int, char **);
static int m_admin(char *, int, char **);
static int m_lusers(char *, int, char **);
static int m_stats(char *, int, char **);
static int m_trace(char *, int, char **);
static int m_info(char *, int, char **);
static int m_whois(char *, int, char **);
#ifdef SPLITSERV
static int m_invite(char *, int, char **);
#endif
#if defined(IRCD_HYB7) || defined(IRCD_HYBSVS)
static int m_eob(char *, int, char **);
#endif
static int m_error(char *, int, char **);
#if defined(IRCD_BAHAMUT) || defined(IRCD_HYBSVS)
static int m_svsnick(char *, int, char **);
static int m_svsmode(char *, int, char **);
static int m_svskill(char *, int, char **);
#ifndef IRCD_HYBSVS
static int m_burst(char *, int, char **);
#endif
#endif
#if defined(SUPPORT_HOSTSERV)
static int m_svshost(char *, int, char **);
static int m_svsident(char *, int, char **);
#endif

static void ircd_initialize(void);
static void add_msg_to_hash_table(char *, struct MsgTable *);
static struct MsgTable *msg_hash[MSG_TABLE_SIZE];

/* Order does not matter now, the command will have to be added in
** capital letters due to a "limitation" of the hash table
*/
static struct MsgTable msg_table[] =
{
	{ "PING",       (long) 0, (long) 0, NULL,       m_ping },
	{ "PRIVMSG",    (long) 0, (long) 0, NULL,       m_privmsg },
	{ "NICK",       (long) 0, (long) 0, NULL,       m_nick },
	{ "MODE",       (long) 0, (long) 0, NULL,       m_mode },
	{ "QUIT",       (long) 0, (long) 0, NULL,       m_quit },
#if defined(IRCD_HYB7) || defined(IRCD_BAHAMUT) || defined(IRCD_HYBSVS)
        { "AWAY",       (long) 0, (long) 0, NULL,       m_away },
#endif
        { "KILL",       (long) 0, (long) 0, NULL,       m_kill },
        { "SERVER",     (long) 0, (long) 0, NULL,       m_server },
        { "SQUIT",      (long) 0, (long) 0, NULL,       m_squit },
#ifdef TRACK_VERSION
        { "351",        (long) 0, (long) 0, NULL,       m_351 },
#endif
#ifdef TRACK_ADMIN
	{ "257",        (long) 0, (long) 0, NULL,       m_257 },
	{ "258",        (long) 0, (long) 0, NULL,       m_258 },
	{ "259",        (long) 0, (long) 0, NULL,       m_259 },
#endif
	{ "NOTICE",     (long) 0, (long) 0, NULL,       m_notice },
#ifdef IRCD_2_10
	{ "SERVICE",	(long) 0, (long) 0, NULL,	m_service },
#endif
	{ "VERSION",	(long) 0, (long) 0, NULL,	m_version },
	{ "MOTD",	(long) 0, (long) 0, NULL,	m_motd },
	{ "ADMIN",	(long) 0, (long) 0, NULL,	m_admin },
	{ "LUSERS",	(long) 0, (long) 0, NULL,	m_lusers },
	{ "STATS",	(long) 0, (long) 0, NULL,	m_stats },
#ifdef SPLITSERV
	{ "INVITE",	(long) 0, (long) 0, NULL,	m_invite },
#endif
        { "WHOIS",	(long) 0, (long) 0, NULL,	m_whois },
	{ "TRACE",	(long) 0, (long) 0, NULL,	m_trace },
	{ "TIME",       (long) 0, (long) 0, NULL,       m_time },
	{ "INFO",	(long) 0, (long) 0, NULL,	m_info },
#if defined(IRCD_HYB7) || defined(IRCD_HYBSVS)
        { "EOB",        (long) 0, (long) 0, NULL,       m_eob },
#endif
        { "ERROR",	(long) 0, (long) 0, NULL,	m_error },
#if defined(IRCD_BAHAMUT) || defined(IRCD_HYBSVS)
	{ "SVSNICK",	(long) 0, (long) 0, NULL,	m_svsnick },
	{ "SVSMODE",	(long) 0, (long) 0, NULL,	m_svsmode },
	{ "SVSKILL",	(long) 0, (long) 0, NULL,	m_svskill },
#ifndef IRCD_HYBSVS
	{ "BURST",	(long) 0, (long) 0, NULL,	m_burst },
#endif
#endif
#if defined(SUPPORT_HOSTSERV)
        { "SVSHOST",    (long) 0, (long) 0, NULL,       m_svshost },
        { "CHGHOST",    (long) 0, (long) 0, NULL,       m_svshost },
        { "SETHOST",    (long) 0, (long) 0, NULL,       m_svshost },
        { "SVSIDENT",   (long) 0, (long) 0, NULL,       m_svsident },
        { "CHGIDENT",   (long) 0, (long) 0, NULL,       m_svsident },
        { "SETIDENT",   (long) 0, (long) 0, NULL,       m_svsident },
#endif
        { (char *)0,    (long) 0, (long) 0, NULL,	(int (*)()) 0 }
};

static void add_msg_to_hash_table(char *msg, struct MsgTable *m)
{
	char *workmsg = strdup(msg);
	strupper(workmsg);
        m->hash = HASH((unsigned char *) workmsg, MSG_TABLE_SIZE);
        m->next = msg_hash[m->hash];
        msg_hash[m->hash] = m;
	free(workmsg);
}

void InitMsgHash(void)
{
	int i;

	for (i = 0; msg_table[i].handler != NULL; i++)
	{
		add_msg_to_hash_table(msg_table[i].name, &msg_table[i]);
	}
}

void parse(char *line)
{
	char *origin = NULL;
	char *pos;
        char *command = NULL;
        char *message = NULL;
        char *parv[MAXPARA];
	static char coreLine[BUFSIZE];
        int parc = 0;
        int i;
	int status = -1;
	struct MsgTable *hashent;

	/* Clear the parv */
        for (i = 0; i < MAXPARA; i++)
                parv[i] = NULL;

	/* Clear out the error string each run through parse() */
	errstr[0] = '\0';

	if (line)
	{
                /* Sometimes we'll get a blank line with just a \n on it...
                ** catch those here....they'll core us later on if we don't
                */
                if (*line == '\n')
                        return;
                if (*line == '\000')
                        return;

	        /* For debugging purposes... */
		strlcpy(coreLine, line, BUFSIZE);

        	debug("R: %s", line);
#ifdef IO_LOG
		iolog("R: %s", line);
#endif

		/* Find the first space */
		if ((pos = strchr(line, ' ')))
		{
			*pos = '\0';
                        pos++;
                        /* If it starts with a :...we have a prefix/origin
                        ** pull the origin off into origin, and have pos for
                        ** the command, message will be the part afterwards
                        */
                        if (*line == ':')
                        {
                                origin = line;
                                origin++;
                                if ((message = strchr(pos, ' ')))
				{
                                        *message = '\0';
                                        message++;
                                        command = pos;
                                }
				else
				{
					command = pos;
					message = NULL;
				}
                        }
			else
			{
                                        message = pos;
                                        command = line;
                        }
                }
                /* Ok, all the nasty part is over, now we need to make a
                ** parv out of what's left
                */

                if (message)
		{
                        if (*message == ':')
			{
                                message++;
                                parv[0] = message;
                                parc = 1;
                        }
			else
			{
                                parc = tokenize(message, parv);
                        }
		}
		else
		{
                        parc = 0;
                }

                /* Now we should have origin (or NULL), command, and a parv
                ** with it's accompanying parc...lets make ABSOLUTELY sure
                */
		if (!command)
		{
			log("Command not found: %s", coreLine);
			return;
		}

		/* Take the command through the hash table */
		strupper(command);
		hashent = msg_hash[HASH((unsigned char *) command,
					 MSG_TABLE_SIZE)];
		while ((hashent != NULL) &&
				(strcmp(hashent->name, command) != 0))
			hashent = hashent->next;
		if (hashent != NULL)
                {
                    status = hashent->handler(origin, parc, parv);
		    hashent->counter++;
		    if (status == 1)
			log("Error with command %s: %s", command,
				errstr ? errstr : "No error message");
                }
	}
	else
	{
		/* Something is REALLY weird here...we got a line that's
		** NULL...for now...just return;
		*/
	}
}

/* All handlers accept the origin as the first parameter (or NULL if there
** is no prefix.  Next is a parameter count, and a parameter vector
** Each handler returns 0 on success or 1 on failure
*/

static int m_time(char *origin, int parc, char *parv[])
{
	if (origin)
	{
		spy("TIME Request by %s", origin);
		sendto_dcc_flag(CF_COMMANDS, "TIME Request by %s", origin);
		sts(":%s %d %s %s :%s", settings.name, RPL_TIME, origin,
			settings.name, irc_time(time(NULL)));
		return 0;
	}

	strcpy(errstr, "No Origin for TIME");
	return 1;
}

static int m_version(char *origin, int parc, char *parv[])
{
	if (origin)
	{
		spy("VERSION Request by %s", origin);
		sendto_dcc_flag(CF_COMMANDS, "VERSION Request by %s", origin);
		sts(":%s %d %s %s :%s %s -> %s", settings.name,
			RPL_VERSION, origin, stats_version,
			settings.name, VersionFlags(), stats_date);
		return 0;
	}

	strcpy(errstr, "No Origin for VERSION");
	return 1;
}

static int m_whois(char *origin, int parc, char *parv[])
{
	if (parc < 2)
	{
		strcpy(errstr, "No Target for WHOIS");
		return 1;
	}
	if (origin)
	{
		struct User *u=finduser(origin);

		spy("Remote WHOIS Request by %s for %s", origin, parv[1]);
		sendto_dcc_flag(CF_COMMANDS,
			"Remote WHOIS Request by %s for %s", origin, parv[1]);
                if (!strcasecmp(parv[1], AdminServ.nick))
                {
                        /* Here is where to answer the WHOIS query */
                        sts(":%s %d %s %s %s %s * :%s", settings.name,
                                RPL_WHOISUSER, u->nick, AdminServ.nick,
                                AdminServ.user, AdminServ.host,
				"Administration Service");
                        sts(":%s %d %s %s %s :%s", settings.name,
                                RPL_WHOISSERVER, u->nick, AdminServ.nick,
                                settings.name, settings.desc);
			/* Always true for AdminServ */
                        sts(":%s %d %s %s :is an IRC Operator",
                                settings.name, RPL_WHOISOPERATOR,
                                u->nick, AdminServ.nick);
                        sts(":%s %d %s %s %d %ld :seconds idle, signon time",
                                settings.name, RPL_WHOISIDLE, u->nick,
                                AdminServ.nick, 0, me.t_start);
                        sts(":%s %d %s %s :End of WHOIS", settings.name,
                                RPL_ENDOFWHOIS, u->nick, AdminServ.nick);
                        return 0;
                }
#ifdef JUPESERV
                if (!strcasecmp(parv[1], JupeServ.nick))
                {
                        /* Here is where to answer the WHOIS query */
                        sts(":%s %d %s %s %s %s * :%s", settings.name,
                                RPL_WHOISUSER, u->nick, JupeServ.nick,
                                JupeServ.user, JupeServ.host,
                                "Jupe Service");
                        sts(":%s %d %s %s %s :%s", settings.name,
                                RPL_WHOISSERVER, u->nick, JupeServ.nick,
                                settings.name, settings.desc);
                        /* Always true for JupeServ */
                        sts(":%s %d %s %s :is an IRC Operator",
                                settings.name, RPL_WHOISOPERATOR,
                                u->nick, JupeServ.nick);
                        sts(":%s %d %s %s %d %ld :seconds idle, signon time",
                                settings.name, RPL_WHOISIDLE, u->nick,
                                JupeServ.nick, 0, me.t_start);
                        sts(":%s %d %s %s :End of WHOIS", settings.name,
                                RPL_ENDOFWHOIS, u->nick, JupeServ.nick);
                        return 0;
                }
#endif
#ifdef STATSERV
		if (!strcasecmp(parv[1], StatServ.nick))
		{
			/* Here is where to answer the WHOIS query */
			sts(":%s %d %s %s %s %s * :%s", settings.name,
				RPL_WHOISUSER, u->nick, StatServ.nick,
				StatServ.user, StatServ.host, StatServ.rname);
			sts(":%s %d %s %s %s :%s", settings.name,
				RPL_WHOISSERVER, u->nick, StatServ.nick,
				settings.name, settings.desc);
			if (IsSetOper())
			{
				sts(":%s %d %s %s :is an IRC Operator",
					settings.name, RPL_WHOISOPERATOR,
					u->nick, StatServ.nick);
			}
			sts(":%s %d %s %s %d %ld :seconds idle, signon time",
				settings.name, RPL_WHOISIDLE, u->nick,
				StatServ.nick, 0, me.t_start);
			sts(":%s %d %s %s :End of WHOIS", settings.name,
				RPL_ENDOFWHOIS, u->nick, StatServ.nick);
			return 0;
		}
#endif
#ifdef SPLITSERV
                if (!strcasecmp(parv[1], SplitServ.nick))
		{
                        /* Here is where to answer the WHOIS query */
                        sts(":%s %d %s %s %s %s * :%s", settings.name,
                                RPL_WHOISUSER, u->nick, SplitServ.nick,
                                SplitServ.user, SplitServ.host, sp.desc);
                        sts(":%s %d %s %s %s :%s", settings.name,
                                RPL_WHOISSERVER, u->nick, SplitServ.nick,
                                settings.name, settings.desc);
                        sts(":%s %d %s %s %d %ld :seconds idle, signon time",
                                settings.name, RPL_WHOISIDLE, u->nick,
                                SplitServ.nick, 0, me.t_start);
                        sts(":%s %d %s %s :End of WHOIS", settings.name,
                                RPL_ENDOFWHOIS, u->nick, SplitServ.nick);
                        return 0;
                }
#endif
		/* Don't answer any others */
		return 0;
	}

	strcpy(errstr, "No Origin for WHOIS");
	return 1;
}

static int m_info(char *origin, int parc, char *parv[])
{
	if (origin)
	{
		const char **text = infotext;
		spy("INFO Request by %s", origin);
		sendto_dcc_flag(CF_COMMANDS, "INFO Request by %s", origin);
		while (*text)
		{
	                if (**text)
        	                sts(":%s %d %s :%s", settings.name, RPL_INFO,
					origin, *text);
                	else
                        	sts(":%s %d %s : ", settings.name, RPL_INFO,
					origin);
                	text++;
		}
		sts("%s %d %s :End of /INFO list.", settings.name,
			RPL_ENDOFINFO, origin);
		return 0;
        }

        strcpy(errstr, "No Origin for INFO");
        return 1;
}

static int m_motd(char *origin, int parc, char *parv[])
{
	if (origin)
	{
		spy("MOTD Request by %s", origin);
		sendto_dcc_flag(CF_COMMANDS, "MOTD Request by %s", origin);
		ShowMOTD(origin);
		return 0;
	}

	strcpy(errstr, "No Origin for MOTD");
	return 1;
}

static int m_privmsg(char *origin, int parc, char *parv[])
{
	if (origin)
	{
		/* parv[0] is the target of the PRIVMSG
		** parv[1] is the complete message
		*/

		/* If these don't exist, we have problems */
		if (parc < 2)
		{
			strcpy(errstr, "Missing parameter in PRIVMSG");
			return 1;
		}

#ifdef STATSERV
#ifdef DISALLOW_NICK_PRIVMSG
		if (!strcasecmp(parv[0], StatServ.nick))
		{
			privmsg(origin, StatServ.nick, MSG_DEFAULT,
				"Please use /msg %s@%s message",
				StatServ.user, settings.name);
			privmsg(origin, StatServ.nick, MSG_DEFAULT,
				"Message not delivered.");
			return 0;
		}
#else
		/* First check for PRIVATE messages to the Stats client */
		if (!strcasecmp(parv[0], StatServ.nick))
		{
#ifdef IRCD_2_10
			statserv(origin, parv[0], parv[1], 0);
#else
			statserv(origin, parv[0], parv[1]);
#endif
			return 0;
		}
#endif
#endif /* STATSERV */

		/* This checks for channel CTCP messages */
		if ( is_channel(parv[0]) )
		{
#ifdef STATSERV
			if (parv[1][0] == '\001')
			{
#ifdef IRCD_2_10
				statserv(origin, parv[0], parv[1], 0);
#else
				statserv(origin, parv[0], parv[1]);
#endif
			}
#endif
			/* Channel messages are ignored...but are NOT errors */
#ifdef SPLITSERV
			/* Unless they are in the Split service's channel */
	                if (!strcasecmp(parv[0], sp.join_chan)
#ifndef SPLITSERV_CUSTOM
			    && (parv[1][0] == '!')
#endif
			    )
			{
	                        splitserv(origin, parv[0], parv[1]);
      		                return 0;
                	}
#endif
			return 0;
		}

#ifdef SPLITSERV
                if (!strcasecmp(parv[0], SplitServ.nick))
		{
                        splitserv(origin, parv[0], parv[1]);
                        return 0;
                }
#endif

		if (!strcasecmp(parv[0], AdminServ.nick))
		{
			adminserv(origin, parv[1]);
			return 0;
		}

#ifdef JUPESERV
		if (!strcasecmp(parv[0], JupeServ.nick))
		{
			jupeserv(origin, parv[1]);
			return 0;
		}
#endif

#ifdef SUPPORT_MSG_UH
#ifdef STATSERV
		if (!strcasecmp(parv[0], s_UH))
		{
#ifdef IRCD_2_10
			statserv(origin, parv[0], parv[1], 0);
#else
			statserv(origin, parv[0], parv[1]);
#endif
			return 0;
		}
#endif
#ifdef SPLITSERV
		if (!strcasecmp(parv[0], sp_UH))
		{
                        splitserv(origin, parv[0], parv[1]);
                        return 0;
                }
#endif
                if (!strcasecmp(parv[0], a_UH))
		{
                        adminserv(origin, parv[1]);
                        return 0;
                }
#ifdef JUPESERV
		if (!strcasecmp(parv[0], j_UH))
		{
			jupeserv(origin, parv[1]);
			return 0;
		}
#endif
#endif
#ifdef SUPPORT_MSG_OPERS
		/* This will allow PRIVMSG opers@server in Hybrid 6 and (soon
		** to be fixed... Hybrid 7) to work.
		*/
		if (!strcasecmp(parv[0], s_OPERS))
		{
			MsgOpers(origin, parv[1]);
			return 0;
		}
#endif
	}

	/* We should NOT get here... The only way a PRIVMSG to
	** something on the service can get here is we used QUOTE
	** to create the user.  We will ignore those messages.
	*/

	snprintf(errstr, BUFSIZE, "Unknown PRIVMSG target: %s", parv[0]);
	return 1;
}

static int m_notice(char *origin, int parc, char *parv[])
{
	if (parc < 2)
	{
		strcpy(errstr, "Missing parameters in NOTICE");
		return 1;
	}
	/* Check for the NOTICE AUTH :... at the beginning of a connection.
	** It's normal and is NOT an error. NOTICE AUTH doesn't have an
	** origin (on hybrid...) so the check for an origin must be below
	** this check.
	*/
	if (!strcasecmp(parv[0], "AUTH"))
		return 0;
	if (!origin)
	{
		strcpy(errstr, "NOTICE without an origin");
		return 1;
	}
#ifdef SUPPORT_MSG_OPERS
	if (!strcasecmp(parv[0], s_OPERS))
	{
		MsgOpers(origin, parv[1]);
		return 0;
	}
#endif
	snprintf(errstr, BUFSIZE, "Unknown NOTICE: %s - %s - %s", origin, 
		parv[0], parv[1]);
	return 1;
}

static int m_server(char *origin, int parc, char *parv[])
{
	/* Hybrid based servers:  parv[0] = server name, parv[1] = hops,
	** parv[2] = description, with or without an origin
	** IRCnet with an origin: parv[0] = server name, parv[1] = hops,
	** parv[2] = server token, parv[3] = description, Without an origin
	** IRCnet acts like the rest.
	*/
	if (parc < 3)
	{
		strcpy(errstr, "Missing parameters in SERVER");
		return 1;
	}
	if (origin)
	{
#ifdef SPLITSERV
		time_t now = time(NULL);
		struct SpServer *spl = NULL;
		char *td = NULL;
#endif
#ifdef IRCD_2_10
		if (parc < 4)
		{
			strcpy(errstr, "Missing parameter in SERVER");
			return 1;
		}
		AddServer(parv[0], atoi(parv[1]), atoi(parv[2]), parv[3],
			origin);
#else
		AddServer(parv[0], atoi(parv[1]), parv[2], origin);
#endif
#ifdef SPLITSERV
		spl = SpFindSplit(parv[0]);
		if (spl)
		{
#ifdef TRACK_SPLITS_ON_DISCONNECT
			/* Sorry, if we're in delayed mode and
			** TRACK_SPLITS_ON_DISCONNECT is enabled, do
			** not print out the notices
			*/
			if (!is_delayed)
			{
#endif
				td = timediff(now - spl->ts);
				splog("End of Split: %s to %s: %s", parv[0],
					origin, td);
				if (IsSPReport())
				{
				  struct Server *serv = findserver(parv[0]);
				  struct Server *sorig = findserver(origin);
				  if (IsSPHide())
				  {
				    if (!IsSHide(serv))
				    {
				      privmsg(sp.join_chan, SplitServ.nick,
					  MSG_SPDEFAULT,
					  "END OF SPLIT: %s to %s: %s", parv[0],
					  IsSHide(sorig) ? "HIDDEN" : origin,
					  td);
				    }
				  }
				  else
				  {
				    privmsg(sp.join_chan, SplitServ.nick,
					MSG_SPDEFAULT,
                                        "END OF SPLIT: %s to %s: %s", parv[0],
					origin, td);
				  }
				}
#ifdef TRACK_SPLITS_ON_DISCONNECT
			}
#endif
		}
		else
		{
			if (!is_delayed)
			{
				splog("New Server: %s to %s", parv[0], origin);
				if (IsSPReport())
				{
				  struct Server *serv = findserver(parv[0]);
				  struct Server *sorig = findserver(origin);
                                  if (IsSPHide())
                                  {
                                    if (!IsSHide(serv))
                                    {
                                      privmsg(sp.join_chan, SplitServ.nick,
                                          MSG_SPDEFAULT,
                                          "CONNECT: %s to %s", parv[0],
                                          IsSHide(sorig) ? "HIDDEN" : origin);
                                    }
                                  }
                                  else
                                  {
                                    privmsg(sp.join_chan, SplitServ.nick,
                                        MSG_SPDEFAULT,
                                        "CONNECT: %s to %s", parv[0], origin);
                                  }
				}
			}
		}
		SpNewServer(parv[0], origin);
		SpDelSplit(parv[0]);
#endif
		return 0;
	}
	/* Now parse SERVER without an origin (our uplink) */

        /* Once we receive the SERVER line, we have connected to our uplink
        ** sucessfully.  Now initialize all clients/channels/etc
        */
        ircd_initialize();

#ifdef IRCD_2_10
		/* The uplink is always token 1 */
	AddServer(parv[0], atoi(parv[1]), 1, parv[2], settings.name);
#else
	AddServer(parv[0], atoi(parv[1]), parv[2], settings.name);
#endif
	/* This is our uplink, so set me.s */
	me.s = findserver(parv[0]);

#ifdef SPLITSERV
	splog("New Server: %s", parv[0]);
	if (!is_delayed)
	{
		if (IsSPReport())
		{
		  struct Server *serv = findserver(parv[0]);
		  struct Server *sorig = findserver(settings.name);
                  if (IsSPHide())
                  {
                    if (!IsSHide(serv))
                    {
                      privmsg(sp.join_chan, SplitServ.nick,
                          MSG_SPDEFAULT,
                          "CONNECT: %s to %s", parv[0],
                          IsSHide(sorig) ? "HIDDEN" : origin);
                    }
                  }
                  else
                  {
                    privmsg(sp.join_chan, SplitServ.nick,
                        MSG_SPDEFAULT,
                        "CONNECT: %s to %s", parv[0], origin);
                  }
		}
	}
	SpNewServer(parv[0], settings.name);
	SpDelSplit(parv[0]);
	map_uplink=SpFindServer(parv[0]);
#endif
	return 0;
}

static int m_squit(char *origin, int parc, char *parv[])
{
#ifdef JUPESERV
	struct Jupe *j = FindJupe(parv[0]);
#endif

	/* Check if there's any case where an SQUIT is prefixed and not
	** from a user, if not, then we can do an if (origin) and show the
	** remote SQUIT notice if the fake server was introduced by a QUOTE
	*/

	DelServer(parv[0]);

#ifdef JUPESERV
	if (j)
	{
		if (settings.flags & FLAG_JUPE_PERSIST)
		{
			alert("%s: JUPE Reintroduced After SQUIT: %s",
				JupeServ.nick, parv[0]);
			sendto_dcc_flag(CF_JUPE,
				"%s: JUPE Reintroduced After SQUIT: %s",
				JupeServ.nick, parv[0]);
			InitJupe(j);
		}
		else
		{
			alert("%s: JUPE Removed by SQUIT: %s", JupeServ.nick,
				parv[0]);
			sendto_dcc_flag(CF_JUPE,
				"%s: JUPE Removed by SQUIT: %s", JupeServ.nick,
				parv[0]);
			DelJupe(parv[0]);
			SaveJupes();
		}
		/* Send a hybrid like remote SQUIT wallops */
		wallops(settings.name, "Remote SQUIT %s from %s (%s)",
			parv[0], origin, parv[1] ? parv[1] : "No Reason");
	}
#endif

	return 0;
}

static int m_quit(char *origin, int parc, char *parv[])
{
	if (origin)
	{
#ifdef IRCD_2_10
		char *host = NULL;
		if ((host = strchr(origin, '@')))
		{
			*host = '\0';
			host++;
			spy("Service %s@%s QUIT: %s", origin, host,
				(parv[0]) ? parv[0] : "NO QUIT MESSAGE");
			sendto_dcc_flag(CF_SERVICES, "Service %s@%s QUIT: %s",
				origin, host,
				(parv[0]) ? parv[0] : "NO QUIT MESSAGE");
			DelService(origin, host);
			return 0;
		}
#endif
		spy("User %s QUIT: %s", origin,
			(parv[0]) ? parv[0] : "NO QUIT MESSAGE");
		sendto_dcc_flag(CF_USER, "User %s QUIT: %s", origin,
			(parv[0]) ? parv[0] : "NO QUIT MESSAGE");
		DelUser(origin);
		return 0;
	}

	strcpy(errstr, "QUIT without an origin");
	return 1;
}

#ifdef TRACK_VERSION
static int m_351(char *origin, int parc, char *parv[])
{
	/* parv[0] = the target (who sent the VERSION query)
	** parv[1], [2], and [3] = The three parts of the VERSION reply
	*/
	struct Server *s = NULL;
	if (origin)
	{
		s = findserver(origin);
		if (s)
		{
			snprintf(s->version, BUFSIZE, "%s %s %s", parv[1],
				parv[2] ? parv[2] : s->name,
				parv[3] ? parv[3] : "");
			return 0;
		}
	}

	strcpy(errstr, "351 Numeric without an origin");
	return 1;
}
#endif

#ifdef TRACK_ADMIN
static int m_257(char *origin, int parc, char *parv[])
{
        /* parv[0] = the target (who sent the ADMIN query)
        ** parv[1] = The ADMIN reply
        */
        struct Server *s = NULL;
        if (origin)
        {
                s = findserver(origin);
                if (s)
                {
			strlcpy(s->admin1, parv[1], LINE);
                        return 0;
                }
        }

        strcpy(errstr, "257 Numeric without an origin");
        return 1;
}

static int m_258(char *origin, int parc, char *parv[])
{
        /* parv[0] = the target (who sent the ADMIN query)
        ** parv[1] = The ADMIN reply
        */
        struct Server *s = NULL;
        if (origin)
        {
                s = findserver(origin);
                if (s)
                {
                        strlcpy(s->admin2, parv[1], LINE);
                        return 0;
                }
        }

        strcpy(errstr, "258 Numeric without an origin");
        return 1;
}

static int m_259(char *origin, int parc, char *parv[])
{
        /* parv[0] = the target (who sent the ADMIN query)
        ** parv[1] = The ADMIN reply
        */
        struct Server *s = NULL;
        if (origin)
        {
                s = findserver(origin);
                if (s)
                {
                        strlcpy(s->admin3, parv[1], LINE);
                        return 0;
                }
        }

        strcpy(errstr, "259 Numeric without an origin");
        return 1;
}
#endif

static int m_mode(char *origin, int parc, char *parv[])
{
	if (origin)
	{
		if (parc < 2)
		{
			strcpy(errstr, "Missing parameters in MODE");
			return 1;
		}
		if ( !is_channel(parv[0]) )
			UserMode(origin, parv[1]);
		return 0;
	}

	strcpy(errstr, "MODE without an origin");
	return 1;
}

#ifdef SPLITSERV
static int m_invite(char *origin, int parc, char *parv[])
{
	if (origin)
	{
		spy("Invite Request by %s to %s channel %s", origin, parv[0],
			parv[1]);
		sendto_dcc_flag(CF_COMMANDS,
			"Invite Request by %s to %s channel %s", origin,
			parv[0], parv[1]);
		if (!strcasecmp(parv[0], SplitServ.nick) &&
			!strcasecmp(parv[1], sp.join_chan))
		{
			init_spjoin();
		}

		return 0;
	}

	strcpy(errstr, "INVITE without an origin");
	return 1;
}
#endif

#if defined(IRCD_HYB7) || defined(IRCD_HYBSVS)
static int m_eob(char *origin, int parc, char *parv[])
{
	if (me.s != NULL)
	{
		if (strcasecmp(me.s->name, origin) == 0)
		{
			SetEOB(me.s);
			me.s->eob_time = time(NULL);
			if (IsNotifyEOB())
			{
				wallops(settings.name, "Burst from the uplink "
					"completed: %d seconds",
					me.s->eob_time - me.s->connected_since);
			}
			return 0;
		}
		strcpy(errstr, "EOB from a server other than the origin");
		return 1;
	}
	strcpy(errstr, "EOB received without being connected to a server");
	return 1;
}
#endif

static int m_error(char *origin, int parc, char *parv[])
{
	if (parv[0] != NULL)
	{
		log("ERROR received from server: %s", parv[0]);
		return 0;
	}
	strcpy(errstr, "NULL ERROR received from server");
	return 1;
}

static int m_kill(char *origin, int parc, char *parv[])
{
	struct User *u = NULL;
	struct Server *s = NULL;

	if (parc < 1)
	{
		strcpy(errstr, "No target in KILL");
		return 1;
	}
	if (origin)
	{
#ifdef IRCD_2_10
                char *host = NULL;
                if ((host = strchr(parv[0], '@')))
                {
                        *host = '\0';
                        host++;
                        spy("Service %s@%s KILLed by %s : %s", parv[0], host,
                                origin,(parv[1]) ? parv[1] : "NO KILL MESSAGE");
			sendto_dcc_flag(CF_SERVICES,
				"Service %s@%s KILLed by %s : %s", parv[0],
				host, origin,(parv[1]) ? parv[1] :
				"NO KILL MESSAGE");
                        DelService(parv[0], host);
                        u = finduser(origin);
                        if (u)
                        {
                                u->server->stats->operkills++;
                        }
                        else
                        {
                                s = findserver(origin);
                                if (s)
                                        s->stats->serverkills++;
                        }
                        return 0;
                }
#endif
		u = finduser(origin);
		spy("Kill Request by %s for %s path: %s", origin, parv[0],
			(parv[1]) ? parv[1] : "NO PATH/REASON");
		sendto_dcc_flag(CF_KILL, "Kill Request by %s for %s path: %s",
			origin, parv[0], (parv[1]) ? parv[1] :
			"NO PATH/REASON");
		if (u)
		{
			u->server->stats->operkills++;
		}
		else
		{
			s = findserver(origin);
			if (s)
				s->stats->serverkills++;
		}
		DelUser(parv[0]);
		if (!strcasecmp(parv[0], AdminServ.nick))
		{
#if defined(SPYMODE) || defined(ALERTMODE)
			ad_delayed = 1;
			eventAdd("check admin delayed", evh_CheckAdDelayed,
				NULL,DEFAULT_DELAY);
#endif
                        init_admin();
#ifdef SPYMODE
			init_spy_join();
#endif
#ifdef ALERTMODE
			init_alert_join();
#endif

			HandleCollide();
#ifdef STATSERV
		}
		else if (!strcasecmp(parv[0], StatServ.nick))
		{
			init_statserv();
			init_join();
			HandleCollide();
#endif
#ifdef SPLITSERV
		}
		else if (!strcasecmp(parv[0], SplitServ.nick))
		{
			init_splitserv();
			init_spjoin();
			HandleCollide();
#endif
#ifdef JUPESERV
		}
		else if (!strcasecmp(parv[0], JupeServ.nick))
		{
			init_jupeserv();
			HandleCollide();
#endif
		}
		return 0;
	}

	strcpy(errstr, "KILL without an origin");
	return 1;
}

#if defined(IRCD_HYB7) || defined(IRCD_BAHAMUT) || defined(IRCD_HYBSVS)
static int m_away(char *origin, int parc, char *parv[])
{
	/* Hybrid 5 and 6 do not propagate AWAY.  IRCnet does, but via the
	** +a usermode
	*/
	struct User *u = NULL;
	if (origin)
	{
		u = finduser(origin);
		if (u)
		{
			/* This might need to be done better */
			if (IsAway(u))
			{
				netinfo.away--;
				u->flags &= ~UFLAG_ISAWAY;
				if (u->server != NULL)
				{
					u->server->away--;
				}
			}
			else
			{
				u->flags |= UFLAG_ISAWAY;
				IncreaseAway(u->server);
			}
			return 0;
		}
	}

	strcpy(errstr, "AWAY without an origin");
	return 1;
}
#endif

static int m_nick(char *origin, int parc, char *parv[])
{
	struct User *u = NULL;
#ifdef IRCD_2_10
	struct Server *s = NULL;
#endif
	if (origin)
	{
		if (parc < 1)
			return 1;
		/* With an origin, it's a nick change 
		** NOTE:  IRCnet puts a : in front of the new nick, while
		** The others don't.  It doesn't matter with the new parser
		** as it will properly deal with it.  The old one didn't
		*/

		/* Incorrect :(
		** csircd prefixes NICK lines, violating TSora, it has to
		** be handles because it.
		** But... since we can use parc, we can save some cycles
		*/
		if (parc <= 2)
		{
			/* 2 would be the new nick and the TS */
			if ( (u = finduser(origin)) )
			{
#if defined(IRCD_BAHAMUT) || defined(IRCD_HYBSVS)
				Change_User(u, parv[0], 0);
#else
				Change_User(u, parv[0]);
#endif
				return 0;
			}
			snprintf(errstr, BUFSIZE, "Unknown user: %s", origin);
			return 1;
		}
	}
	/* Without an origin, it's a nick introduction...
	** parv contents vary for each server type...
	** Hybrid:  0 = nick, 1 = hops, 2 = TS, 3 = modes, 4 = user
	**          5 = host, 6 = server, 7 = gecos
	** IRCnet: 0 = nick, 1 = hops, 2 = user, 3 = host, 4 = server token,
	**         5 = modes, 6 = gecos
	** Bahamut:  0 = nick, 1 = hops, 2 = TS, 3 = modes, 4 = user
	**           5 = host, 6 = server, 7 = servicestamp, 8 = gecos
	*/

	if (parc < 7)
	{
		strcpy(errstr, "Missing parameter in NICK introduction");
		return 1;
	}
#if !defined(IRCD_2_10) && !defined(IRCD_BAHAMUT)
	if (parc < 8)
	{
		strcpy(errstr, "Missing parameter in NICK introduction");
		return 1;
	}
#endif

#if defined(IRCD_HYBRID) || defined(IRCD_HYB6) || defined(IRCD_HYB7) || \
    defined(IRCD_HYBSVS)
	spy("User Connect %s from %s@%s on %s gecos: %s", parv[0], parv[4],
		parv[5], parv[6], parv[parc - 1]);
	sendto_dcc_flag(CF_USER, "User Connect %s from %s@%s on %s gecos: %s",
		parv[0], parv[4], parv[5], parv[6], parv[parc - 1]);
	AddUser(parv[0], parv[4], parv[5], parv[6]);
	UserMode(parv[0], parv[3]);
#elif defined(IRCD_2_10)
	s = findservtoken(atoi(parv[4]));
	spy("User Connect %s from %s@%s on %s gecos: %s", parv[0], parv[2],
		parv[3], s->name, parv[parc - 1]);
	sendto_dcc_flag(CF_USER, "User Connect %s from %s@%s on %s gecos: %s",
		parv[0], parv[2], parv[3], s->name, parv[parc - 1]);
	AddUser(parv[0], parv[2], parv[3], s);
	UserMode(parv[0], parv[5]);
#elif defined(IRCD_BAHAMUT)
	spy("User Connect %s from %s@%s on %s gecos: %s", parv[0], parv[4],
		parv[5], parv[6], parv[parc - 1]);
	sendto_dcc_flag(CF_USER, "User Connect %s from %s@%s on %s gecos: %s",
		parv[0], parv[4], parv[5], parv[6], parv[parc - 1]);
	AddUser(parv[0], parv[4], parv[5], parv[6]);
	UserMode(parv[0], parv[3]);
#endif
	return 0;
}

#ifdef IRCD_2_10
static int m_service(char *origin, int parc, char *parv[])
{
	struct Service *sv = NULL;
	char *host = NULL;

	host = strchr(parv[0], '@');
	if (host)
	{
		*host = '\0';
		host ++;
	}
	else
	{
		/* Something weird is going on... */
		strcpy(errstr, "SERVICE received without @");
		return 1;
	}

	sv = findservice(parv[0], host);

	if (sv)
	{
		strcpy(errstr, "Reintroduced Service!");
		return 1;
	}

	/* Parameters:  parv[0] (after processing) = nick, host = server,
	**              parv[1] = hopcount, parv[2] = distribution,
	**		parv[3] = parameters, parv[4] = server token,
	**		parv[5] = gecos
	*/

	spy("Service Connect %s@%s dist %s mode %s gecos: %s", parv[0], host,
		parv[2], parv[3], parv[5]);
	sendto_dcc_flag(CF_SERVICES,
		"Service Connect %s@%s dist %s mode %s gecos: %s", parv[0],
		host, parv[2], parv[3], parv[5]);
	AddService(parv[0], host, parv[2], parv[5]);

	return 0;
}
#endif

static int m_ping(char *origin, int parc, char *parv[])
{
	if (origin)
	{
		/* It's a remote Lag detecting PING */
		if (parc < 2)
		{
			strcpy(errstr, "Missing parameter in PING");
			return 1;
		}
		/* Lag detecting PING's we reverse the order of the two
		** parameters
		*/
		sts(":%s PONG %s :%s", settings.name, parv[1], parv[0]);
		return 0;
	}
	/* Without an origin, it's our uplink checking us...be sure to
	** answer or we'll ping timeout...this PING is like a heartbeat 
	** to the service as well.
	*/

	if (parc < 1) {
		strcpy(errstr, "Missing parameter in PING");
		return 1;
	}
	sts("PONG %s", parv[0]);
	return 0;
}

static int m_admin(char *origin, int parc, char *parv[])
{
	if (origin)
	{
		/* Is it directed to us or a jupe behind us? */
		/* We don't support jupe, but one can be made with
		** quote/raw
		*/
		if (parv[0])
		{
			spy("ADMIN Request by %s to %s", origin, parv[0]);
			sendto_dcc_flag(CF_COMMANDS,
				"ADMIN Request by %s to %s", origin,
				parv[0]);
			/* We should NEVER get an ADMIN query that doesn't
			** have a valid parameter
			*/
			if (!strcasecmp(settings.name, parv[0]))
			{
			    sts(":%s %d %s :Administrative info about %s",
				settings.name, RPL_ADMINME, origin,
				settings.name);
			    sts(":%s %d %s :%s Network Statistical Services",
				settings.name, RPL_ADMINLOC1, origin,
				settings.netname);
			    sts(":%s %d %s :%s", settings.name, RPL_ADMINLOC2,
				origin, settings.admin_name);
			    sts(":%s %d %s :%s", settings.name, RPL_ADMINEMAIL,
				origin, settings.admin_email);
			    return 0;
			}

			snprintf(errstr, BUFSIZE,
				"Received ADMIN request for unknown leaf: %s",
				parv[0]);
			return 1;
		}
		strcpy(errstr, "Missing parameter in ADMIN");
		return 1;
	}
	strcpy(errstr, "ADMIN without an origin");
	return 1;
}

static int m_lusers(char *origin, int parc, char *parv[])
{
        if (origin)
        {
                /* Is it directed to us or a jupe behind us? */
                /* We don't support jupe, but one can be made with
                ** quote/raw
                */
                if (parv[1])
                {
                        spy("LUSERS Request by %s to %s for %s", origin,
				parv[1], parv[0]);
			sendto_dcc_flag(CF_COMMANDS,
				"LUSERS Request by %s to %s for %s", origin,
				parv[1], parv[0]);
                        /* We should NEVER get an LUSERS query that doesn't
                        ** have a valid parameter
                        */
			/* Ignore the mask option...it's not totally correct
			** but why bother
			*/
                        if (!strcasecmp(settings.name, parv[1]))
                        {
			    int count = 1; /* AdminServ */
#ifdef SPLITSERV
			    count++;
#endif
#ifdef STATSERV
			    count++;
#endif
#ifdef JUPESERV
			    count++;
#endif
#ifdef IRCD_2_10
			    sts(":%s %d %s :There are %ld users and %d services"
				" on %d servers", settings.name,
				RPL_LUSERCLIENT, origin, netinfo.users,
				netinfo.services, netinfo.servers);
#else
			    /* Track +i users count */
			    sts(":%s %d %s :There are %ld users and 0 invisible"
				" on %d servers", settings.name,
				RPL_LUSERCLIENT, origin, netinfo.users,
				netinfo.servers);
#endif
			    sts(":%s %d %s %d :IRC Operators online",
				settings.name, RPL_LUSEROP, origin,
				netinfo.opers);
			    /* I would send numeric 254 RPL_LUSERCHANNELS
			    ** here once I track channels correctly
			    */
#ifdef IRCD_2_10
			    sts(":%s %d %s :I have %d users, 0 services and 1 "
				"servers", settings.name, RPL_LUSERME, origin,
				count);
#else
			    sts(":%s %d %s :I have %d clients and 1 servers",
				settings.name, RPL_LUSERME, origin, count);
#endif
			    /* I need to send these as some statistical clients
			    ** parse these
			    */
			    sts(":%s %d %s :Current local  users: %d Max: %d",
				settings.name, RPL_LOCALUSERS, origin, count,
				count);
			    sts(":%s %d %s :Current global users: %ld Max: %ld",
				settings.name, RPL_GLOBALUSERS, origin,
				netinfo.users, netinfo.maxusers);
                            return 0;
                        }

                        snprintf(errstr, BUFSIZE,
                                "Received LUSERS request for unknown leaf: %s",
                                parv[0]);
                        return 1;
                }
                strcpy(errstr, "Missing parameter in LUSERS");
                return 1;
        }
        strcpy(errstr, "LUSERS without an origin");
        return 1;
}

static int m_stats(char *origin, int parc, char *parv[])
{
	/* parv[0] = type, parv[1] = target */
	struct User *u = NULL;
	if (parc < 2)
	{
		strcpy(errstr, "Missing parameter in STATS");
		return 1;
	}
	if (origin)
	{
		u = finduser(origin);
		spy("STATS %c Request by %s to %s", parv[0][0], origin,
			parv[1]);
		sendto_dcc_flag(CF_COMMANDS, "STATS %c Request by %s to %s",
			parv[0][0], origin, parv[1]);
		if (u)
		{
			if (!strcasecmp(settings.name, parv[1]))
			{
				ParseStats(u, parv[0][0]);
				return 0;
			}
		}
		snprintf(errstr, BUFSIZE, "Unknown user: %s in STATS", origin);
		return 1;
	}
	strcpy(errstr, "No origin in STATS");
	return 1;
}

static int m_trace(char *origin, int parc, char *parv[])
{
	struct User *u = NULL;
	if (parc < 1)
	{
		strcpy(errstr, "Missing parameter in TRACE");
		return 1;
	}
	if (origin)
	{
		u = finduser(origin);
		spy("TRACE Request by %s for %s", origin, parv[0]);
		sendto_dcc_flag(CF_COMMANDS, "TRACE Request by %s for %s",
			origin, parv[0]);
		if (u)
		{
			if (!strcasecmp(settings.name, parv[0]))
			{
				Trace(u);
				return 0;
			}
		}
		snprintf(errstr, BUFSIZE, "Unknown user: %s in TRACE", origin);
		return 1;
	}
	strcpy(errstr, "No origin in TRACE");
	return 1;
}

#if defined(SUPPORT_HOSTSERV)
static int m_svshost(char *origin, int parc, char *parv[])
{
       struct User *u;

        if (origin)
        {
               if (parc < 2)
               {
                       strcpy(errstr, "Missing Parameters in SVSHOST");
                       return 1;
               }
               if ( (u = finduser(parv[0])) )
               {
                       Change_Host(u, parv[1]);
                       return 0;
               }
               snprintf(errstr, BUFSIZE, "Unkown user: %s", parv[0]);
               return 1;
       }
       strcpy(errstr, "No Origin in SVSHOST");
       return 1;
}

static int m_svsident(char *origin, int parc, char *parv[])
{
        struct User *u;
                 
        if (origin)
        {
                if (parc < 2)
                {
                        strcpy(errstr, "Missing Parameters in SVSDENT");
                        return 1;
                }
                if ( (u = finduser(parv[0])) )
                {
                        Change_Ident(u, parv[1]);
                        return 0;
                }   
                snprintf(errstr, BUFSIZE, "Unkown user: %s", parv[0]);
                return 1;
        }
        strcpy(errstr, "No Origin in SVSIDENT");
        return 1;
}
#endif

#if defined(IRCD_BAHAMUT) || defined(IRCD_HYBSVS)
static int m_svsnick(char *origin, int parc, char *parv[])
{
	struct User *u;

	if (origin)
	{
		if (parc < 3)
		{
			strcpy(errstr, "Missing Parameters in SVSNICK");
			return 1;
		}
		/* XXX Track stats on origin */
		if ( (u = finduser(parv[0])) )
		{
			Change_User(u, parv[1], 1);
			return 0;
		}
		snprintf(errstr, BUFSIZE, "Unknown user: %s", parv[0]);
		return 1;
	}
	strcpy(errstr, "No Origin in SVSNICK");
	return 1;
}

static int m_svsmode(char *origin, int parc, char *parv[])
{
	/* Old format:  0 = nick, 1 = mode, 2 = servicesID
	** New format:  0 = nick, 1 = ts, 2 = mode, 3 = optional servicesID
	*/
	char *modes;
        if (origin)
        {
                if (parc < 3)
                {
                        strcpy(errstr, "Missing parameters in MODE");
                        return 1;
                }
		/* SVSMODE has incompatabilities... */
		if ((parv[1][0] == '+') || (parv[1][0] == '-'))
			modes = parv[1];
		else
			modes = parv[2];
                if ( !is_channel(parv[0]) )
                        UserMode(parv[0], modes);
                return 0;
        }

        strcpy(errstr, "SVSMODE without an origin");
        return 1;
}

static int m_svskill(char *origin, int parc, char *parv[])
{
        if (parc < 2)
        {
                strcpy(errstr, "No target in SVSKILL");
                return 1;
        }
        if (origin)
        {
                spy("SvsKill Request by %s for %s path: %s", origin, parv[0],
                        (parv[2]) ? parv[2] : "NO REASON");
		sendto_dcc_flag(CF_SERVICES,
			"SvsKill Request by %s for %s path: %s", origin,
			parv[0], (parv[2]) ? parv[2] : "NO REASON");
		/* XXX - Track SVS in server structs */
                DelUser(parv[0]);
                if (!strcasecmp(parv[0], AdminServ.nick))
                {
#if defined(SPYMODE) || defined(ALERTMODE)
                        ad_delayed = 1;
                        eventAdd("check admin delayed", evh_CheckAdDelayed,
                                NULL,DEFAULT_DELAY);
#endif
                        init_admin();
#ifdef SPYMODE
                        init_spy_join();
#endif
#ifdef ALERTMODE
                        init_alert_join();
#endif

                        HandleCollide();
#ifdef STATSERV
                }
                else if (!strcasecmp(parv[0], StatServ.nick))
                {
                        init_statserv();
                        init_join();
                        HandleCollide();
#endif
#ifdef SPLITSERV
                }
                else if (!strcasecmp(parv[0], SplitServ.nick))
                {
                        init_splitserv();
                        init_spjoin();
                        HandleCollide();
#endif
#ifdef JUPESERV
                }
                else if (!strcasecmp(parv[0], JupeServ.nick))
                {
                        init_jupeserv();
                        HandleCollide();
#endif
                }
                return 0;
        }

        strcpy(errstr, "SVSKILL without an origin");
        return 1;
}

/* XXX - Should I track SVSNOOP here at all? */

#ifndef IRCD_HYBSVS
static int m_burst(char *origin, int parc, char *parv[])
{
	if (parc < 1)
	{
		/* Start of burst, we don't care about that
		**
		** but maybe we should?
		*/
		return 0;
	}
        if (me.s != NULL)
        {
                if (strcasecmp(me.s->name, origin) == 0)
                {
                        SetEOB(me.s);
                        me.s->eob_time = time(NULL);
                        if (IsNotifyEOB())
                        {
                                wallops(settings.name, "Burst from the uplink "
                                        "completed: %d seconds",
                                        me.s->eob_time - me.s->connected_since);
                        }
                        return 0;
                }
                strcpy(errstr, "BURST from a server other than the origin");
                return 1;
        }
        strcpy(errstr, "BURST received without being connected to a server");
        return 1;
}
#endif /* IRCD_HYBSVS */

#endif /* IRCD_BAHAMUT */

static void ircd_initialize(void)
{
#ifdef IRCD_BAHAMUT
	sts(":%s BURST", settings.name);
#endif
#ifdef STATSERV
	init_statserv();
#endif
#ifdef SPLITSERV
	init_splitserv();
	init_spjoin();
#endif
	init_admin();
#ifdef JUPESERV
	init_jupeserv();
	FlushJupeList();
	LoadJupes();
#endif
#ifdef STATSERV
	init_join();
#endif
#ifdef SPYMODE
	init_spy_join();
#endif
#ifdef ALERTMODE
	init_alert_join();
#endif
#if defined(IRCD_HYB7) || defined(IRCD_HYBSVS)
	sts(":%s EOB", settings.name);
#endif
#ifdef IRCD_BAHAMUT
	sts(":%s BURST 0", settings.name);
#endif
}

/* Send a message to a user from a user */
void privmsg(char *to, const char *from, int type, char *fmt, ...)
{
	va_list ap;
	char buf[BUFSIZE], buf2[BUFSIZE];
	char *cmd;

	/* Since privmsg can send a wide variety of messages,
	** determine what type here
	*/
	switch (type)
	{
		case MSG_DEFAULT:
			if (IsPrivmsg())
				cmd = "PRIVMSG";
			else
				cmd = "NOTICE";
			break;
#ifdef SPLITSERV
		case MSG_SPDEFAULT:
			if (is_channel(to))
			{
				if (IsSPChanmsg())
					cmd = "PRIVMSG";
				else
					cmd = "NOTICE";
			}
			else
			{
				if (IsSPPrivmsg())
					cmd = "PRIVMSG";
                                else
                                        cmd = "NOTICE";
                        }
			break;
#endif
		case MSG_PRIVMSG:
			cmd = "PRIVMSG";
			break;
		case MSG_NOTICE:
			cmd = "NOTICE";
			break;
		case MSG_SQUERY:
			cmd = "SQUERY";
			break;
		default:
			/* Default to PRIVMSG */
			cmd = "PRIVMSG";
			break;
	}

	va_start(ap, fmt);
	vsnprintf(buf2, sizeof(buf2), fmt, ap);
	snprintf(buf, BUFSIZE, ":%s %s %s :%s", from, cmd, to, buf2);
	sts("%s", buf);
#ifdef DELAY_PRIVMSG
	select_delay();
#endif
	va_end(ap);
}

/* Read a help file and send it to the recipient via privmsg() */
void send_help_file(char *to, char *from, int type, char *filename, char *query)
{
	FILE *fp = fopen(filename, "r");
	char buffer[BUFSIZE];

	if (!fp)
	{
		privmsg(to, from, type, "No help available for: %s", query);
		log("No help available for %s, filename %s", query, filename);
		return;
	}

	while (fgets(buffer, BUFSIZE, fp))
	{
		strip(buffer);
		if (buffer[0])
			privmsg(to, from, type, "%s", buffer);
		else
			privmsg(to, from, type, " ");
	}

	fclose(fp);
}

void wallops(char *from, char *fmt, ...)
{
	va_list ap;
	char buf[BUFSIZE], buf2[BUFSIZE];

	if (IsSilent())
		return;

	va_start(ap, fmt);
	vsnprintf(buf2, sizeof(buf2), fmt, ap);
	snprintf(buf, BUFSIZE, ":%s WALLOPS :%s", from, buf2);
	sts("%s", buf);
	va_end(ap);
}

#ifdef STATSERV
/* This function KILL's or optionally ignores a user for flooding services. */
/* This needs to be adjusted for SplitServ too! */
int flood(struct User *u)
{
	/* Our "opers" are immune to flood control */
	if (u->myuser) 
		return 0;

	/* Flood control is disabled by setting the flood lines to 0 */
	if (settings.flood)
	{
		time_t current = time(NULL);

		if (current - u->t_flood > settings.flood_time)
		{
			u->t_flood = current;
			u->flood = 0;
			u->flags &= ~UFLAG_FLOODSENT;
			return 0;
		}

		if (u->flood >= settings.flood)
		{
		   /* This was a switch in the initial writing of this
		   ** For now (unless I see a specific reason against
		   ** using switch, it can stay a switch)
		   */
		   if (!IsFlood())
		   {
			sts(":%s KILL %s :%s!%s (Flooding Services.)",
				StatServ.nick, u->nick, StatServ.host, 
				StatServ.nick);
			log("FLOODING (KILLED): %s!%s@%s", u->nick, 
				u->username, u->hostname);
			/* This is an unnecessary piece of overhead ugh */
			/* We KILL the user, we should remove them from 
			** the databases...
			*/
			DelUser(u->nick);
		   }
		   else
		   {
			/* Only send the flood WALLOPS/GLOBOPS message
			** once, or WE will be the ones flooding
			*/
			if (!IsFloodSent(u)) {
			  wallops("%s is Flooding Services", u->nick);
			  log("FLOODING (IGNORED): %s!%s@%s", u->nick,
				  u->username, u->hostname);
			}
		   }
		   return 1;
		}
		else
		{
			u->flood++;
		}
	}
	return 0;
}
#endif

/* This is left in unmodified as to the original author's request */
static void ShowMOTD(char *nick)
{
	static unsigned char copyright[38] = {  /* Can you tell I was bored? */
	0x47, 0x65, 0x6f, 0x53, 0x74, 0x61, 0x74, 0x73, 0x20, 0x61, 0x72, 0x65,
	0x20, 0x28, 0x63, 0x29, 0x20, 0x31, 0x39, 0x39, 0x39, 0x20, 0x4a, 0x6f,
	0x6e, 0x61, 0x74, 0x68, 0x61, 0x6e, 0x20, 0x47, 0x65, 0x6f, 0x72, 0x67,
	0x65, 0x0
	};

	/* Removal or modification of this function is against the law.
	** Be nice, and don't remove it or modify it.
	** -Jonathan 'netgod' George (net@lite.net) */

	sts(":%s 375 %s :- %s Message of the Day", settings.name, nick, settings.name);
	sts(":%s 372 %s :- ", settings.name, nick);
	sts(":%s 372 %s :- %s", settings.name, nick, copyright);
	sts(":%s 372 %s :- Grab your copy at ftp.lite.net", settings.name, nick);
	sts(":%s 376 %s :End of /MOTD command.", settings.name, nick);
}

/* Send out QUIT messages as a courtesy... */
void SendQuit(struct User *u, int restart, char * comm)
{
#ifdef STATSERV
	sts(":%s QUIT :%s Request by %s: %s", StatServ.nick, 
	    (restart) ? "Restart" : "Shutdown", u->nick,
            (comm && comm[0]) ? comm : "No Reason");
#endif
#ifdef SPLITSERV
	sts(":%s QUIT :%s Request by %s: %s", SplitServ.nick,
	    (restart) ? "Restart" : "Shutdown", u->nick,
	    (comm && comm[0]) ? comm : "No Reason");
#endif
#ifdef JUPESERV
	sts(":%s QUIT :%s Request by %s: %s", JupeServ.nick,
	    (restart) ? "Restart" : "Shutdown", u->nick,
            (comm && comm[0]) ? comm : "No Reason");
#endif
	sts(":%s QUIT :%s Request by %s: %s", AdminServ.nick,
            (restart) ? "Restart" : "Shutdown", u->nick,
            (comm && comm[0]) ? comm : "No Reason");
}

/* This is how we answer STATS requests...
** This is very Hybrid-like at the moment.  It should not affect other
** ircd's.  If someone can refine this for other ircd's, I will patch it in
*/
static void ParseStats(struct User *u, char type)
{
    struct MyUser *myu = NULL;
    int count = 0;
    int i = 0;

    switch(type)
    {
	case 'C':
	case 'c':
		/* Our only "C and N" lines are to our uplink, show the
		** real IP/address of our uplink to opers, but use the
		** Hybrid6 method of IP hiding for non-opers.  This is
		** moot if your ircd blocks remote STATS from non-opers
		*/
		sts(":%s %d %s C %s A %s %d uplink", settings.name,
			RPL_STATSCLINE, u->nick, 
#ifdef HIDE_IPS
			"*@127.0.0.1",
#else
			IsOper(u) ? settings.uplink : "*@127.0.0.1", 
#endif
			me.s->name, settings.port);
		/* Removed N lines */
		break;
	case 'E':
	case 'e':
		/* Only show events to logged in users */
		if (u->myuser != NULL)
			show_events(u->nick);
		break;
	case 'I':
	case 'i':
		/* This is purely cosmetic, show a closed server */
		sts(":%s %d %s I * * *@%s 0 nonopered", settings.name,
			RPL_STATSILINE, u->nick, settings.name);
		break;
	case 'H':
	case 'h':
		/* Our only hub should be our uplink.  This is cosmetic
		** as well.
		*/
		sts(":%s %d %s H * * %s", settings.name, RPL_STATSHLINE,
			u->nick, me.s->name);
		break;
	case 'M':
	case 'm':
		/* Show the known commands and their usage counters */
		for (i=0; msg_table[i].handler; i++)
		{
		    /* lie about the byte count */
		    sts(":%s %d %s %s %ld 0", settings.name, RPL_STATSCOMMANDS,
			u->nick, msg_table[i].name, msg_table[i].counter);
		}
		break;
	case 'O':
	case 'o':
	    /* This STATS is actually very useful.  This shows a list of
	    ** our users (all who can use LOGIN, the equivalent of opers)
	    */
	    for (myu = myuhead; myu; myu = myu->next)
	    {
		sts(":%s %d %s O *@* * %s 0 opered", settings.name,
			RPL_STATSOLINE, u->nick, myu->username);
	    }
            break;
	/* P is ports, we don't listen (yet) so ignore it */
	case 'p':
	    /* This STATS is also relatively useful.  It does the Hybrid
	    ** STATS p - show all online opers.  All users who have used
	    ** LOGIN will be shown as Stats opers.
	    ** NOTE:  We now DO print out the current user's nick.
	    */
	    for (myu = myuhead; myu; myu = myu->next)
	    {
		if (IsOperOn(myu))
		{
		    sts(":%s %d %s :[O] %s", settings.name, RPL_STATSDEBUG,
			u->nick, myu->u->nick);
		    count++;
		}
		/* XXX - They CAN be online in both places...which one should
		** be shown, or just keep it showing both
		*/
		if (IsOperDCC(myu))
		{
		    sts(":%s %d %s :[D] %s", settings.name, RPL_STATSDEBUG,
		    u->nick, myu->conn->name);
		    count++;
		}
	    }
	    sts(":%s %d %s :%d OPER%s", settings.name, RPL_STATSDEBUG, u->nick,
		count, (count == 1) ? "" : "s");
	    break;
	/* U is U: lines/shared blocks, ignore */
	case 'u':
		/* Send our uptime in the usual fasion */
		sts(":%s %d %s :Statistics up %s", settings.name,
			RPL_STATSUPTIME, u->nick,
			timediff(time(NULL) - me.t_start));
		break;
	case 'T':
	case 't':
		/* For now, just show the byte counters */
		sts(":%s %d %s :Bytes in: %ld, Bytes out: %ld", settings.name,
			RPL_STATSDEBUG, u->nick, netinfo.inb, netinfo.outb);
                break;
	case 'Y':
	case 'y':
		/* Purely cosmetic with fake values...EXCEPT for the 
		** reconnect time in the server Y line (2)
		*/
		sts(":%s %d %s Y nonopered 0 0 2 0", settings.name,
			RPL_STATSYLINE, u->nick);
		sts(":%s %d %s Y uplink 0 %d 1 0", settings.name, 
			RPL_STATSYLINE, u->nick, settings.r_time);
		sts(":%s %d %s Y opered 0 0 1 0", settings.name, 
			RPL_STATSYLINE, u->nick);
		break;
	default:
                break;
    }
    sts(":%s %d %s %c :End of STATS report", settings.name,
		RPL_ENDOFSTATS, u->nick, type);
}

/* Report TRACE in the Hybrid style */
static void Trace(struct User *u)
{
    if (IsOper(u))
    {
#ifdef STATSERV
	   sts(":%s %d %s %s %s %s[%s@%s] (127.0.0.1) 0 0",
		settings.name, 
		IsSetOper() ? RPL_TRACEOPERATOR : RPL_TRACEUSER, 
		u->nick, 
                IsSetOper() ? "Oper" : "User", 
		IsSetOper() ? "opered" : "nonopered", StatServ.nick,
		StatServ.user, StatServ.host);
#endif
#ifdef SPLITSERV
	   sts(":%s %d %s User nonopered %s[%s@%s] (127.0.0.1) 0 0",
		settings.name, RPL_TRACEUSER, u->nick, SplitServ.nick,
		SplitServ.user, SplitServ.host);
#endif
	   sts(":%s %d %s Oper opered %s[%s@%s] (127.0.0.1) 0 0",
		settings.name, RPL_TRACEOPERATOR, u->nick, AdminServ.nick,
		AdminServ.user, AdminServ.host);
#ifdef JUPESERV
	   sts(":%s %d %s Oper opered %s[%s@%s] (127.0.0.1) 0 0",
		settings.name, RPL_TRACEOPERATOR, u->nick, JupeServ.nick,
		JupeServ.user, JupeServ.host);
#endif
           sts(":%s %d %s Serv uplink %dS %ldC %s AutoConn.!*@%s 0",
                    settings.name, RPL_TRACESERVER, u->nick, netinfo.maxservers, 
		    netinfo.maxusers, me.s->name, settings.name);
	   sts(":%s %d %s %s :End of TRACE", settings.name, RPL_ENDOFTRACE, u->nick, 
		    settings.name);
    }
}

#ifdef SUPPORT_MSG_OPERS
static void MsgOpers(char *nick, char *message)
{
	struct MyUser* myu = NULL;

	for (myu = myuhead; myu; myu = myu->next)
	{
		if (IsOperOn(myu))
			sts(":%s NOTICE %s :To opers: From %s: %s",
				settings.name, myu->u->nick, nick, message);
		if (IsOperDCC(myu))
			sendto_dcc(myu->conn->fd, "To opers: From %s: %s",
				nick, message);
        }
}
#endif

static void CollideDie()
{
                wallops(settings.name, "Collision Flood!  Gracefully Exiting");
                log("Collision Flood!  Gracefully Exiting");
#ifdef STATSERV
                SendQuit(finduser(StatServ.nick), 0, "Collision Flood");
#endif
#ifdef SPLITSERV
		SendQuit(finduser(SplitServ.nick), 0, "Collision Flood");
#endif
#ifdef JUPESERV
		SendQuit(finduser(JupeServ.nick), 0, "Collision Flood");
#endif
		SendQuit(finduser(AdminServ.nick), 0, "Collision Flood");
                sleep(1);
                close(servsock);
		close_dccs();
                SaveStats();
                exit(0);
}

static void HandleCollide()
{
#ifdef DIE_ON_MASS_COLLIDE
        netinfo.collides++;
        if (netinfo.collides > MAXCOLLISIONS)
        {
                CollideDie();
        }
#endif
}
