/* Stats - IRC Statistical Services for Hybrid, csircd, and IRCnet 2.10
** servers.c - Functions to track server state
**
** Copyright W. Campbell and others.  See README for more details
** Some code Copyright: Jonathan George, Kai Seidler, ircd-hybrid Team,
**                      IRCnet IRCD developers.
**
** $Id: servers.c,v 1.9 2002/10/29 03:17:02 wcampbel Exp $
*/

#include <assert.h>
#include "stats.h"
#include "struct.h"
#include "proto.h"
#include "match.h"

static struct Server *new_server(char *, char *);
struct Server *serverlist[S_TABLE_SIZE];

static void add_server_to_hash_table(char *, struct Server *);
static void del_server_from_hash_table(struct Server *);

static void add_server_to_hash_table(char *name, struct Server *s)
{
	s->hash = HASH((unsigned char *) name, S_TABLE_SIZE);
	s->next = serverlist[s->hash];
	serverlist[s->hash] = (void *)s;
        debug("add_server_to_hash_table: %s, %d", name, s->hash);
}

static void del_server_from_hash_table(struct Server *s)
{
	struct Server *tmp, *prev = NULL;

	for (tmp = serverlist[s->hash]; tmp; tmp = tmp->next)
	{
		if (tmp == s)
		{
			if (prev)
				prev->next = tmp->next;
			else
				serverlist[s->hash] = tmp->next;
			tmp->next = NULL;
			return;
		}
		prev = tmp;
	}
}

static struct Server *new_server(char *name, char *desc)
{
	struct Server *s;

	s = smalloc(sizeof(struct Server));
	if (!name)
		name = "";
	strlcpy(s->name, name, MAXHOST);
	if (!desc)
		desc = "";
#if defined(IRCD_HYBRID) || defined(IRCD_HYB6) || defined(IRCD_HYB7) || \
    defined(IRCD_HYBSVS)
	/* Remove the (H) hidden identifier */
	if ( (desc[0] == '(') &&
             (desc[1] == 'H') &&
             (desc[2] == ')') )
		desc += 4;
#endif
	strlcpy(s->desc, desc, LINE);
	add_server_to_hash_table(name, s);

	debug("new_server: %s, %s", name, desc);

	return s;
}

#ifdef IRCD_2_10
void AddServer(char *name, int hops, int token, char *desc, char* uplink)
#else
void AddServer(char *name, int hops, char *desc, char* uplink)
#endif
{
	struct Server *s;
	time_t current = time(NULL);

#ifdef JUPESERV
#if defined(TRACK_ADMIN) || defined(TRACK_VERSION)
	struct Jupe *j = FindJupe(name);
#endif
#endif

	debug("New Server: %s", name);
	strlower(name);

	/* Check to see if a server already exists - Why is a mystery to me 
	** but sometimes on Hyb7 it'll happen
	*/
	s = findserver(name);
	if (s)
	{
		log("Server %s already exists in the server table", name);
	}
	else
	{
		s = new_server(name, desc);
	}
	s->hops = hops;
#ifdef IRCD_2_10
	s->token = token;
#endif
	s->connected_since = current;
	s->last_announce = current;
        if (uplink != NULL)
                s->uplink = findserver(uplink);

	debug("findserver() --> %s uplink %s", s->name,
		s->uplink ? s->uplink->name : "NULL");

#ifdef TRACK_VERSION
	/* If it is NOT settings.name */
	if (strcasecmp(settings.name, name))
	{
#ifdef JUPESERV
		if (j == NULL)
#endif
			sts(":%s VERSION %s", AdminServ.nick, name);
	}
	else
	{
		snprintf(s->version, BUFSIZE, "%s %s %s -> %s",
			stats_version, settings.name, VersionFlags(),
			stats_date);
	}
#endif

#ifdef TRACK_ADMIN
        /* If it is NOT settings.name */
        if (strcasecmp(settings.name, name))
        {
#ifdef JUPESERV
		if (j == NULL)
#endif
                	sts(":%s ADMIN %s", AdminServ.nick, name);
		/* admin items do not need to be zero'ed, calloc'ed memory */
        }
        else
        {
                snprintf(s->admin1, BUFSIZE, "%s Network Statistical Services",
                        settings.netname);
		snprintf(s->admin2, BUFSIZE, "%s", settings.admin_name);
		snprintf(s->admin3, BUFSIZE, "%s", settings.admin_email);
        }
#endif

	/* Check to see if it should be hidden */
	if (findhidden(name))
		s->flags |= SFLAG_HIDE;

	/* Check for the (H) in the server desc field, and mark the
	** server as hidden.  We probably should not add it to the 
	** Hidden list though, it'll be a waste
	*/

#if defined(IRCD_HYBRID) || defined(IRCD_HYB6) || defined(IRCD_HYB7) || \
    defined(IRCD_HYBSVS)
	/* Keep it enabled even if we are not linked to a Hybrid 7
	** server, just in case there's a "hidden hub" on the network
	*/
	if (desc)
		if ( (desc[0] == '(') &&
		     (desc[1] == 'H') &&
                     (desc[2] == ')') )
			s->flags |= SFLAG_HIDE;
#endif
	
	IncreaseServers();
	AddStats(s);
}

/* This is similar to DelServer, but always recurses down the tree!
** This function is used when the connection to the uplink is lost
*/
void FlushServer(char *name)
{
        struct Server *up = NULL;
        struct User* u = NULL;
	char *tstr = NULL;
	struct Server *s = findserver(name);
        int i=0;
	debug("FlushServer called for %s", name);

	if (!s || !name)
	{
		log("FlushServer(): %s failed!", name);
                return;
        }

        u=userlist[i];
        while (TRUE)
	{
                if (i == U_TABLE_SIZE)
		{
                        break;
                }
                if (u)
		{
                        if (!strcmp(name, u->server->name))
			{
                                if (u->next)
				{
                                        /* UGH kludge */
                                        tstr = u->nick;
                                        u = u->next;
                                        DelUser(tstr);
                                        tstr = NULL;
                                        continue;
                                }
				else
				{
                                        DelUser(u->nick);
					i++;
					u = userlist[i];
					continue;
                                }
                        }

                        if (u->next)
			{
                                u = u->next;
                        }
			else
			{
                                i++;
                                u = userlist[i];
                        }
                }
		else
		{
			i++;
			u = userlist[i];
		}
        }
        i = 0;
        up = serverlist[i];
        while (TRUE)
	{
                if (i == S_TABLE_SIZE)
		{
                        break;
                }
                if (up)
		{
                        if (up->uplink)
			{
                                if (!strcmp(name, up->uplink->name))
				{
                                        if (up->next)
					{
                                                tstr = up->name;
                                                up = up->next;
                                                FlushServer(tstr);
                                                tstr = NULL;
                                                continue;
                                        }
					else
					{
                                                FlushServer(up->name);
						i++;
						up = serverlist[i];
						continue;
                                        }
                                }
                        }
                        if (up->next)
			{
                                up = up->next;
                        }
			else
			{
                                i++;
                                up = serverlist[i];
                        }
                }
		else
		{
                        i++;
                        up = serverlist[i];
                }
        }

#ifdef SPLITSERV
	SpDelServer(name);
#ifdef TRACK_SPLITS_ON_DISCONNECT
	if (!IsSPHide() || !(IsSPHide() && IsSHide(s)))
		SpNewSplit(name);
#endif
#endif

        del_server_from_hash_table(s);
        s->stats->lastseen = time(NULL);
        DecreaseServers();
        free(s);
}

void DelServer(char *name)
{
	struct Server *s = findserver(name);

#if defined(IRCD_HYB6) || defined(IRCD_HYB7) || defined(IRCD_2_10) || \
    defined(IRCD_BAHAMUT) || defined(IRCD_HYBSVS)
	/* This is to support the QS capab
	** IRCnet 2.10.x does this too
	** Bahamut does via NOQUIT
	*/
	struct Server *up = NULL;
	struct User* u = NULL;
	char *tstr = NULL;
	int i=0;
	u=userlist[i];
#endif

        if (!s || !name)
        {
                log("DelServer(): %s failed!", name);
                return;
        }

        if (s)
                s->stats->numsplits++;

#ifdef IRCD_2_10
        DelServicesByServer(name);
#endif

#if defined(IRCD_HYB6) || defined(IRCD_HYB7) || defined(IRCD_2_10) || \
    defined(IRCD_BAHAMUT) || defined(IRCD_HYBSVS)
	while (TRUE)
	{
		if (i == U_TABLE_SIZE)
		{
			break;
		}
		if (u)
		{
			if (!strcmp(name, u->server->name))
			{
				if (u->next)
				{
					/* UGH kludge */
					tstr = u->nick;
					u = u->next;
					DelUser(tstr);
					tstr = NULL;
					continue;
				}
				else
				{
					DelUser(u->nick);
					i++;
					u = userlist[i];
					continue;
				}
			}

			if (u->next)
			{
				u = u->next;
			}
			else
			{
				i++;
				u = userlist[i];
			}
                }
		else
		{
                        i++;
                        u = userlist[i];
                }
	}
	i = 0;
	up = serverlist[i];
	while (TRUE)
	{
		if (i == S_TABLE_SIZE)
		{
			break;
		}
		if (up)
		{
			if (up->uplink)
			{
				if (!strcmp(name, up->uplink->name))
				{
					if (up->next)
					{
						tstr = up->name;
						up = up->next;
						DelServer(tstr);
						tstr = NULL;
						continue;
					}
					else
					{
						DelServer(up->name);
						i++;
						up = serverlist[i];
						continue;
					}
				}
			}
			if (up->next)
			{
				up = up->next;
			}
			else
			{
				i++;
				up = serverlist[i];
			}
		}
		else
		{
			i++;
			up = serverlist[i];
		}
	}
#endif

#ifdef SPLITSERV
                        splog("Split: %s from %s", name,
                                (s && (s->uplink)) ? s->uplink->name
                                        : "UNKNOWN");
                        if (!is_delayed)
                        {
                                if (IsSPReport())
				{
                                  if (IsSPHide())
                                  {
                                    if (!IsSHide(s))
                                    {
                                      privmsg(sp.join_chan, SplitServ.nick,
                                          MSG_SPDEFAULT,
                                          "SPLIT: %s from %s", name,
                                          IsSHide(s->uplink) ? "HIDDEN" :
					  (s && (s->uplink)) ? s->uplink->name
					  : "UNKNOWN");
                                    }
                                  }
                                  else
                                  {
                                    privmsg(sp.join_chan, SplitServ.nick,
                                        MSG_SPDEFAULT,
                                        "SPLIT: %s from %s", name,
                                        (s && (s->uplink)) ? s->uplink->name :
					"UNKNOWN");
                                  }
				}
                        }

        SpDelServer(name);
	if (!IsSPHide() || !(IsSPHide() && IsSHide(s)))
        	SpNewSplit(name);
#endif

	del_server_from_hash_table(s);
	s->stats->lastseen = time(NULL);
	DecreaseServers();
	assert(netinfo.servers >= 0);
	free(s);
}

#ifdef IRCD_2_10
/* This needs a serious optimization */
struct Server *findservtoken(int token)
{
        struct Server *s;
        int i;

        for(i=0; i<S_TABLE_SIZE; i++)
	{
                s = serverlist[i];
		while (s && (s->token != token))
			s = s->next;
		if (s)
			return s;
        }

	/* We should never get here */
       	debug("findservtoken(%d) -> %s", token, (s) ? s->name : "NOT FOUND");

	/* I'd rather return an explicit NULL here */
        return NULL;
}
#endif

struct Server *findserver(char *name)
{
	struct Server *s;
	int i;

	strlower(name); 
	debug("findserver(%s)", name);
	s = serverlist[HASH((unsigned char *) name, S_TABLE_SIZE)];
	while (s && (strcasecmp(s->name, name) != 0))
		s = s->next;
	if (s)
		return s;

/* The name in the server table is the pattern (with the *) 
** The name is what is passed into this function (usually full 
** server name)
**
** NOTE:  This also is a performance hit, if there are a number of masked
**        servers or there are a lot of users on that masked server/set of
**        servers hidden by the mask
*/
	for(i=0; i<S_TABLE_SIZE; i++)
	{
		s = serverlist[i];
        	while (s && (match(s->name, name) != 0))
                	s = s->next;
		if (s)
			return s;
	}

	/* We should never get here unless the server doesn't exist */
	debug("findserver(%s) -> %s", name, (s) ? s->name : "NOT FOUND");

/* I'd rather return an explicit NULL here */
	return NULL;
}

void init_server_hash()
{
	int i;
	struct Server *s, *b;

	for (i = 0; i < S_TABLE_SIZE; i++)
	{
		s = serverlist[i];
		while (s)
		{
			b = s->next;
			debug("init_server_hash() deleting %s", s->name);
			free(s);
			s = b;
		}
		serverlist[i] = NULL;
	}
	memset((void *)serverlist, '\0', sizeof(serverlist));
#ifdef IRCD_2_10
	AddServer(settings.name, 0, 0, settings.desc, NULL);
#else
	AddServer(settings.name, 0, settings.desc, NULL);
#endif
}

void IncreaseServers()
{
        time_t current = time(NULL);
        debug("Increasing server count");
        netinfo.servers++;
        if (netinfo.servers > netinfo.maxservers)
        {
                netinfo.maxservers = netinfo.servers;
                netinfo.t_maxservers = current;
        }
        if (netinfo.servers > daily.servers)
        {
                daily.servers = netinfo.servers;
                daily.t_servers = current;
        }
}

