/* Stats - IRC Statistical Services for Hybrid, csircd, and IRCnet 2.10
** sock.c - TCP/IP socket management
**
** Copyright W. Campbell and others.  See README for more details
** Some code Copyright: Jonathan George, Kai Seidler, ircd-hybrid Team,
**                      IRCnet IRCD developers.
**
** $Id: sock.c,v 1.69 2002/10/28 14:32:21 wcampbel Exp $
*/

#include "stats.h"
#include "struct.h"
#include "proto.h"
#include "setup.h"
#include "event.h"

fd_set readfds, writefds, nullfds;
struct Connection *sconn = NULL;

static int Server_receiveFrom(int, char *);
static void Service_processPacket(int, char *, char *);

#ifdef HAVE_GETADDRINFO
static struct addrinfo * gethostinfo(char const *host, int port);

int ConnectTo(char *host, int port)
{
	struct addrinfo *hostres, *res;
	struct addrinfo *bindres = NULL;
	int s = -1;

	hostres = gethostinfo(host, port);
	if (hostres == NULL)
	{
		log("FATAL:  Lookup error in CONNECT_TO, exiting");
		exit(6);
	}
	if (IsVHost())
	{
		bindres = gethostinfo(settings.vhost, 0);
		if (bindres == NULL)
		{
			log("FATAL:  Lookup error in VHOST, exiting");
			exit(6);
		}
	}
	for (res = hostres; res; res = res->ai_next)
	{
		s = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
		if (s < 0)
			continue;
		if (IsVHost())
		{
			if (bind(s, bindres->ai_addr, bindres->ai_addrlen) < 0)
			{
				log("Cannot bind to VHost");
				return (-1);
			}
		}
		if (connect(s, res->ai_addr, res->ai_addrlen) == 0)
			break;
		log("%s", strerror(errno));
		close(s);
	}
	freeaddrinfo(hostres);
	if (res == NULL)
	{
		log("Cannot connect to host");
		return (-1);
	}
	sconn = add_connection(s, "Server Uplink", CONN_SERV);

	return s;
}
#else
/* Initiate the connection to the server, returns the FD */
int ConnectTo(char *host, int port)
{
	struct hostent *hp;
	struct sockaddr_in sa;
	struct in_addr * in;
	int s;

	if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	{
		log("Cannot create socket()");
		return (-1);
	}

        if (IsVHost())
	{
                memset((void *) &sa, '\0', sizeof(sa));
                sa.sin_family = AF_INET;
                if ((hp = gethostbyname (settings.vhost)) == NULL)
		{
			log("Unable to gethostbyname(%s)", settings.vhost);
                        return (-1);
                }
                in = (struct in_addr *) (hp->h_addr_list[0]);
                sa.sin_addr.s_addr = in->s_addr;
                sa.sin_port = 0;
                if (bind(s, (struct sockaddr *) &sa, sizeof(sa)) < 0)
		{
                        close(s);
			log("Unable to bind to VHOST %s", settings.vhost);
                        return (-1);
                }
        }

        if ((hp = gethostbyname (host)) == NULL)
        {
                log("Cannot gethostbyname(%s)", host);
                return (-1);
        }

	memset((void *) &sa, '\0', sizeof(sa));
	sa.sin_family = AF_INET;
	sa.sin_port = htons (port);
	memcpy((void *) &sa.sin_addr, (void *) hp->h_addr, hp->h_length);

	if (connect (s, (struct sockaddr *) &sa, sizeof (sa)) < 0)
	{
		close (s);
		log("Cannot connect()");
		return (-1);
	}

	debug("Connected to server(%s) port(%d)", host, port);

	sconn = add_connection(s, "Server Uplink", CONN_SERV);
	return s;
}
#endif

/* Here is where it all happens, get a line from the server
** and send it off to the parse function.  Also this is where
** the call to the event loop is issued.
*/
void read_loop()
{
	int SelectResult;
	struct timeval TimeOut;
	static char buf[BUFSIZE + 1];
	dlink_node *dl;
	struct Connection *conn;

	while (TRUE)
	{
		memset(buf, '\0', BUFSIZE + 1);
		eventRun();
		FD_ZERO(&readfds);
		FD_ZERO(&writefds);
		TimeOut.tv_sec = 1;
		TimeOut.tv_usec = 0L;
		DLINK_FOREACH(dl, connection_list.head)
		{
			conn = (struct Connection *) dl->data;
			if (conn->flags & CONN_CONNECTING)
			{
			    if (conn->flags & CONN_DCCOUT)
				FD_SET(conn->fd, &writefds);
			    else if (conn->flags & CONN_DCCIN)
				FD_SET(conn->fd, &readfds);
			}
			else
				FD_SET(conn->fd, &readfds);
		}
		if ((SelectResult = select(FD_SETSIZE, &readfds, &writefds,
			&nullfds, &TimeOut)) > 0)
		{
		    DLINK_FOREACH(dl, connection_list.head)
		    {
			conn = (struct Connection *) dl->data;
			if (FD_ISSET(conn->fd, &readfds))
			{
			    if (conn->flags & CONN_CONNECTING)
			    {
				int accept_sock;
				struct sockaddr_in incoming_addr;
				int addrlen = sizeof(struct sockaddr);

				/* clear errno before this begins */
				errno = 0;

				if((accept_sock = accept(conn->fd,
					(struct sockaddr *)&incoming_addr,
					(socklen_t *)&addrlen)) < 0)
				{
				    if (errno == EAGAIN)
					continue;
				    privmsg(conn->name, AdminServ.nick,
					MSG_DEFAULT, "DCC CHAT Failed");
				    close(conn->fd);
				    conn->flags |= CONN_DEAD;
				    delete_connection(conn->fd);
				    continue;
				}

				/* close the listening socket */
				close(conn->fd);

				conn->fd = accept_sock;
				conn->flags &= ~CONN_CONNECTING;
				/* Send out the welcome message */
				sendto_dcc(conn->fd, "Connected to %s: %s",
					AdminServ.nick, settings.name);
				sendto_all_dcc("Client %s connected",
					conn->name);
				if (IsUnauthChat())
					conn->flags |= CF_CHAT;
			    }
			    if (!Server_receiveFrom(conn->fd, buf))
			    {
				if (conn->fd == servsock)
				{
				    log("Lost connection to server.");
				    return;
				}
				else
				{
				    /* One of the DCC sockets died */
				    log("Lost connection to dcc socket %d",
					conn->fd);
				    close(conn->fd);
				    conn->flags |= CONN_DEAD;
				    sendto_all_dcc("Client %s disconnected",
					conn->name);
				    delete_connection(conn->fd);
				}
			    }
			    Service_processPacket(conn->fd, buf, conn->hbuffer);
			}
			if (FD_ISSET(conn->fd, &writefds))
			{
                            if (conn->flags & CONN_CONNECTING)
                            {
                                int ret;

                                /* Send a welcome message, this will catch
                                ** if the connection fails by erroring it out
                                */
                                ret = sendto_dcc(conn->fd,
                                        "Connected to %s: %s",
                                        AdminServ.nick, settings.name);
                                if (ret == 0)
                                {
                                        /* This socket has connected */
                                        conn->flags &= ~CONN_CONNECTING;
                                        /* As long as it returns "valid" */
                                        sendto_all_dcc("Client %s connected",
                                                conn->name);
					if (IsUnauthChat())
						conn->flags |= CF_CHAT;
                                }
			    }
			}
		    }
		}
		else
		{
			/* select returns 0 when the TimeOut is hit */
			if (SelectResult == 0)
			{
				/* simply a timeout, needed for chk() */
			}
			else if ((SelectResult == -1) && (errno == EINTR))
			{
				/* Interrupted system call, we got a signal */
			}
			else
			{
				/* select returned -1, an error condition of
				** some kind (we don't care what) occurred.
				*/
				log("Lost connection to server: %d: %s",
					errno, strerror(errno));
				return;
			}
		}
	}
}

/* These two functions were lifted from oswald's NoteServ */
static int Server_receiveFrom(int fd, char *buffer)
{
    /* This mustn't be BUFSIZE!!! */
    /* unless buffer is BUFSIZE + 1 */
    int n = read(fd, buffer, BUFSIZE);
    buffer[n] = '\0';
    return n;
}

static void Service_processPacket(int fd, char *buffer, char *backing)
{
    char *ptr;
    char buffer2[BUFSIZE * 2];

    while ((ptr = strchr(buffer, '\n')))
    {
        *ptr = '\0';
        if (*(ptr - 1) == '\r')
            *(ptr - 1) = '\0';
        snprintf(buffer2, (BUFSIZE * 2), "%s%s", backing, buffer);
        *backing = '\0';
	netinfo.inb += strlen(buffer2);
	if (fd == servsock)
        	parse(buffer2);
	else
		dcc_parse(fd, buffer2);
        buffer = ptr + 1;
    }
    if (*buffer)
    {
        strlcpy(backing, buffer, BUFSIZE * 2);
    }
}

#ifdef HAVE_GETADDRINFO
/* Stolen from FreeBSD's whois client, modified for Stats by W. Campbell */
static struct addrinfo * gethostinfo(char const *host, int port)
{
        struct addrinfo hints, *res;
        int error;
	char portbuf[6];

        memset(&hints, 0, sizeof(hints));
        hints.ai_flags = 0;
        hints.ai_family = AF_UNSPEC;
        hints.ai_socktype = SOCK_STREAM;
	snprintf(portbuf, 6, "%d", port);
        error = getaddrinfo(host, portbuf, &hints, &res);
        if (error) {
                log("gethostinfo error: %s: %s", host, gai_strerror(error));
                return (NULL);
        }
        return (res);
}
#endif

void close_dccs(void)
{
	dlink_node *dl;
	dlink_node *dtmp;
	struct Connection *conn;

	DLINK_FOREACH_SAFE(dl, dtmp, connection_list.head)
	{
		conn = (struct Connection *) dl->data;
		if (conn->flags & (CONN_DCCIN|CONN_DCCOUT))
		{
			close(conn->fd);
			delete_connection(conn->fd);
		}
	}
}

