/* Stats - IRC Statistical Services for Hybrid, csircd, and IRCnet 2.10
** splitserv.c - Split service interface
**
** Copyright W. Campbell and others.  See README for more details
** Some code Copyright: Jonathan George, Kai Seidler, ircd-hybrid Team,
**                      IRCnet IRCD developers.
**
** $Id: splitserv.c,v 1.41 2002/10/28 14:02:09 wcampbel Exp $
*/

#include "stats.h"
#include "struct.h"
#include "proto.h"
#include "text.h"

#ifdef SPLITSERV

struct SplTable
{
        char* name;
        long counter;
	int channel;  /* 1 if it's allowed in a channel */
        int (*handler)();
};

static char errstr[BUFSIZE];

static int sp_map(char *, char *, int, char **);
static int sp_split(char *, char *, int, char **);
static int sp_dump(char *, char *, int, char **);
static int sp_add(char *, char *, int, char **);
static int sp_forget(char *, char *, int, char **);
static int sp_update(char *, char *, int, char **);
static int sp_help(char *, char *, int, char **);
static int sp_chelp(char *, char *, int, char **);
static int sp_cping(char *, char *, int, char **);
static int sp_cversion(char *, char *, int, char **);
static int sp_triggers(char *, char *, int, char **);

static struct SplTable sp_table[] =
{
	{ TRIGGER_SPLIT,0,      1,      sp_split },
	{ TRIGGER_MAP,  0,      1,      sp_map },
        { "MAP",        0,      0,      sp_map },
        { "SPLIT",      0,      0,      sp_split },
        { TRIGGER_DUMP, 0,      1,      sp_dump },
        { "DUMP",       0,      0,      sp_dump },
        { "FORGET",     0,      0,      sp_forget },
        { "ADD",        0,      0,      sp_add },
        { "UPDATE",     0,      0,      sp_update },
	{ "HELP",	0,	0,	sp_help },
	{ TRIGGER_HELP,	0,	1,	sp_chelp },
	{ TRIGGER_TRIGGERS,0,	1,	sp_triggers },
	{ "TRIGGERS",	0,	0,	sp_triggers },
	{ "\001PING",	0,	1,	sp_cping },
	{ "\001VERSION\001",0,	1,	sp_cversion },
	{ (char *)0,	0,	0,	(int (*)()) 0 }
};

static struct HelpTable help_table[] =
{
        { TRIGGER_MAP,  "help/splitserv/map",            0 },
        { "MAP",        "help/splitserv/map",            0 },
	{ TRIGGER_SPLIT,"help/splitserv/split",          0 },
	{ "SPLIT",	"help/splitserv/split",          0 },
	{ TRIGGER_DUMP,	"help/splitserv/dump",           0 },
	{ "DUMP",	"help/splitserv/dump",           0 },
	{ "FORGET",	"help/splitserv/forget",         1 },
	{ "ADD",	"help/splitserv/add",            1 },
	{ "UPDATE",	"help/splitserv/update",         1 },
        { (char *)0,    (char *)0,                       0 }
};


void splitserv(char *origin, char *target, char *param)
{
	char *parv[MAXPARA];
	int parc = 0;
	struct User *u = NULL;
	int i;
	int status;
	static char buf[BUFSIZE];

        /* Clear the parv */
        for (i = 0; i < MAXPARA; i++)
                parv[i] = NULL;

	/* Clear the error string */
	errstr[0] = '\0';

	strlcpy(buf, param, BUFSIZE);

	if (param)
		parc = msgtoken(param, parv);

	if (origin)
		u = finduser(origin);
	else
		return;

        if (!u)
	{
                log("Unable to finduser %s (splitserv)", origin);
                return;
        }

	netinfo.requests++;

	if (!parc)
		return;

	/* This will generate some extraneous log file entries for
	** messages such as :) and .
	*/
	if (!strcasecmp(target, sp.join_chan) &&
		(isalpha((int) parv[0][0]) || isdigit((int) parv[0][0])))
		return;

	if (parv[0][0] != '\001')
		log("%s received message from %s to %s: %s", SplitServ.nick, 
			origin, target, buf);

	/* All the preliminary junk is done, now we have to go through 
	** the tables and call the handler for that command
	*/
        for (i=0; sp_table[i].handler; i++)
        {
                if (!strcasecmp(parv[0], sp_table[i].name))
                {
		  if (is_channel(target) && !sp_table[i].channel)
			return;
                    status = sp_table[i].handler(u->nick, target, parc, parv);
                    if (status == 1)
                        splog("Error with splitserv command %s: %s", parv[0],
			(errstr) ? errstr : "No error message");
		    /* Matched a value, done, then return */
		    return;
                }
        }
	/* If we get here, we did not get a valid command...
	** Make sure not to say anything to the user if it was a 
	** CTCP that was not matched
	*/
	if (*parv[0] == '\001')
		return;
	/* Don't answer invalid commands if it was a message to a channel */
	if (is_channel(target))
		return;
	privmsg(u->nick, SplitServ.nick, MSG_SPDEFAULT, "Invalid command: %s",
		parv[0]);
	return;
}

/* In all of these functions, parv[0] contains the command, followed
** by the parameters
*/

static int sp_chelp(char *origin, char *target, int parc, char *parv[])
{
	if (strcasecmp(sp.join_chan, target))
	{
		/* This happens to not be a channel message */
		return sp_help(origin, target, parc, parv);
	}
	send_help_file(sp.join_chan, SplitServ.nick, MSG_SPDEFAULT,
		"help/splitserv/pindex", "pindex");
	return 0;
}

static int sp_help(char *origin, char *target, int parc, char *parv[])
{
	struct User *u = finduser(origin);
	int i;

	if (parc < 2)
	{
		send_help_file(u->nick, SplitServ.nick, MSG_SPDEFAULT,
			"help/splitserv/index", "index");
		return 0;
	}

	for (i = 0; help_table[i].filename; i++)
	{
		if (!strcasecmp(parv[1], help_table[i].name))
                {
			if (help_table[i].myuonly)
			{
				if (u->myuser)
					send_help_file(u->nick, SplitServ.nick,
						MSG_SPDEFAULT,
						help_table[i].filename,
						help_table[i].name);
				else
					privmsg(u->nick, SplitServ.nick,
						MSG_SPDEFAULT,
						"Permission Denied.");
			}
			else
			{
				send_help_file(u->nick, SplitServ.nick,
					MSG_SPDEFAULT,
					help_table[i].filename,
					help_table[i].name);
			}

			return 0;
		}
	}

	privmsg(u->nick, SplitServ.nick, MSG_SPDEFAULT,
		"Unknown Help Topic: %s", parv[1]);
	snprintf(errstr, BUFSIZE, "Unknown HELP topic: %s", parv[1]);
	return 1;
}

static int sp_map(char *origin, char *target, int parc, char *parv[])
{
	struct User *u;
	struct SpServer *root;

	if (sp.map_setting == SO_DISABLED)
	{
		privmsg(origin, SplitServ.nick, MSG_SPDEFAULT,
			"The map function is disabled.");
		return 0;
	}

	if (sp.map_setting == SO_OPERONLY) 
	{
		u = finduser(origin);
		if (!IsOper(u))
		{
			privmsg(origin, SplitServ.nick, MSG_SPDEFAULT,
				"The map function is only available to opers.");
			return 0;
		}
	}

	if (parv[1] == NULL)
	{
		root = map_head;
	}
	else
	{
		char *dest = is_channel(target) ? target : origin;

		root = SpFindServer(parv[1]);
		if (root == NULL)
		{
			privmsg(dest, SplitServ.nick, MSG_SPDEFAULT,
				"No such server %s", parv[1]);
			return 0;
		}
		privmsg(dest, SplitServ.nick, MSG_SPDEFAULT,
			"NOTICE:  Starting from server %s", parv[1]);
	}

	if (is_channel(target))
		PrintMap(target, root, 0, NULL);
	else
		PrintMap(origin, root, 0, NULL);
	return 0;
}

static int sp_split(char *origin, char *target, int parc, char *parv[])
{
        struct User *u;

        if (sp.split_setting == SO_DISABLED)
        {
                privmsg(origin, SplitServ.nick, MSG_SPDEFAULT,
                        "The split function is disabled.");
                return 0;
        }

        if (sp.split_setting == SO_OPERONLY)
        {
                u = finduser(origin);
                if (!IsOper(u))
                {
                        privmsg(origin, SplitServ.nick, MSG_SPDEFAULT,
                            "The split function is only available to opers.");
                        return 0;
                }
        }

	if (is_channel(target))
		PrintSplit(target);
	else
		PrintSplit(origin);
	return 0;
}

static int sp_dump(char *origin, char *target, int parc, char *parv[])
{
        struct User *u;

        if (sp.dump_setting == SO_DISABLED)
        {
                privmsg(origin, SplitServ.nick, MSG_SPDEFAULT,
                        "The dump function is disabled.");
                return 0;
        }

        if (sp.dump_setting == SO_OPERONLY)
        {
                u = finduser(origin);
                if (!IsOper(u))
                {
                        privmsg(origin, SplitServ.nick, MSG_SPDEFAULT,
                            "The dump function is only available to opers.");
                        return 0;
                }
        }

	if (is_channel(target))
		PrintDump(target);
	else
		PrintDump(origin);
	return 0;
}

static int sp_forget(char *origin, char *target, int parc, char *parv[])
{
	struct User *u = finduser(origin);
	struct SpServer * s = NULL;
	dlink_node *dl = NULL;

	if ((u != NULL) && (u->myuser != NULL))
	{
		if (!strcasecmp(target, SplitServ.nick))
		{
			if (parc < 2)
			{
				privmsg(origin, SplitServ.nick, MSG_SPDEFAULT,
					"Cannot forget a null server");
				return 0;
			}

			for (dl = split_table.head; dl; dl = dl->next)
			{
				s = (struct SpServer *) dl->data;
				if (!strcasecmp(s->name, parv[1]))
				{
				    privmsg(origin, SplitServ.nick,
					MSG_SPDEFAULT,
					"Removed from the Split Table",
					s->name);
				    dlink_delete(dl, &split_table);
				    privmsg(sp.join_chan, SplitServ.nick,
					MSG_SPDEFAULT,
					"FORGOT SPLIT: %s", s->name);
				    splog("Forgot split %s by %s", s->name,
					origin);
				    free(s);
				    free(dl);
				    return 0;
				}
			}

			privmsg(origin, SplitServ.nick, MSG_SPDEFAULT,
				"%s NOT found in the Split Table", parv[1]);
		}
	}
	return 0;
}

static int sp_add(char *origin, char *target, int parc, char *parv[])
{
	struct User *u = finduser(origin);
	struct SpServer *s = NULL;

        if ((u != NULL) && (u->myuser != NULL))
	{
                if (!strcasecmp(target, SplitServ.nick))
		{
			if (parc < 2)
			{
				privmsg(origin, SplitServ.nick, MSG_SPDEFAULT,
				    "Can't add a split for a null server");
				return 0;
			}

			s = SpFindServer(parv[1]);
			if (s != NULL)
			{
				privmsg(origin, SplitServ.nick, MSG_SPDEFAULT,
				    "Can't add a split for an existing server");
				return 0;
			}
			s = SpFindSplit(parv[1]);
			if (s != NULL)
			{
				privmsg(origin, SplitServ.nick, MSG_SPDEFAULT,
				    "Can't add a split for an existing split");
				return 0;
			}
			SpNewSplit(parv[1]);
			privmsg(sp.join_chan, SplitServ.nick, MSG_SPDEFAULT,
				"ADDED SPLIT: %s", parv[1]);
			splog("Added split %s by %s", parv[1], origin);
                }
        }
        return 0;
}

static int sp_update(char *origin, char *target, int parc, char *parv[])
{
        struct User *u = finduser(origin);

        if ((u != NULL) && (u->myuser != NULL))
	{
                if (!strcasecmp(target, SplitServ.nick))
		{
			if (IsSPEnableText())
			{
				dump_map();
				privmsg(origin, SplitServ.nick, MSG_SPDEFAULT,
					"MAP file updated");
				return 0;
			}
			privmsg(origin, SplitServ.nick, MSG_SPDEFAULT,
				"MAP Text Dump Disabled");
                }
        }
        return 0;
}

static int sp_cversion(char *origin, char *target, int parc, char *parv[])
{
        struct User *u = finduser(origin);

        sts(":%s NOTICE %s :\001VERSION %s %s %s -> %s\001",
	     SplitServ.nick, u->nick, stats_version, settings.name,
	     VersionFlags(), stats_date);
        splog("CTCP VERSION from %s!%s@%s", u->nick, u->username, u->hostname);

        return 0;
}

static int sp_cping(char *origin, char *target, int parc, char *parv[])
{
        struct User *u = finduser(origin);
        char command[BUFSIZE];

	ExpandParv(command, BUFSIZE, 1, parc, parv);
        if (command[0] == '\0')
                return 1;
        sts(":%s NOTICE %s :\001PING %s", SplitServ.nick, u->nick, command);
        log("CTCP PING from %s!%s@%s", u->nick, u->username, u->hostname);

        return 0;
}

static int sp_triggers(char *origin, char *target, int parc, char *parv[])
{
        if (is_channel(target))
		privmsg(target, SplitServ.nick, MSG_SPDEFAULT,
			"Public Triggers: %s %s %s",
			TRIGGER_SPLIT, TRIGGER_MAP, TRIGGER_DUMP);
        else
		privmsg(origin, SplitServ.nick, MSG_SPDEFAULT,
			"Public Triggers: %s %s %s",
			TRIGGER_SPLIT, TRIGGER_MAP, TRIGGER_DUMP);
        return 0;
}

#endif
