/* Stats - IRC Statistical Services for Hybrid, csircd, and IRCnet 2.10
** users.c - Functions to track users state
**
** Copyright W. Campbell and others.  See README for more details
** Some code Copyright: Jonathan George, Kai Seidler, ircd-hybrid Team,
**                      IRCnet IRCD developers.
**
** $Id: users.c,v 1.118 2002/10/29 20:00:52 wcampbel Exp $
*/

#include <assert.h>
#include "stats.h"
#include "struct.h"
#include "proto.h"

/* These are ugly functions, gets right down to the hash tables and
** manipulation of raw data
*/

static struct User *new_user(char *);
static void add_user_to_hash_table(char *, struct User *);
static void del_user_from_hash_table(struct User *);

struct User *userlist[U_TABLE_SIZE];

static void add_user_to_hash_table(char *nick, struct User *u)
{
	u->hash = HASH((unsigned char *) nick, U_TABLE_SIZE);
	u->next = userlist[u->hash];
	userlist[u->hash] = (void *)u;
}

static void del_user_from_hash_table(struct User *u)
{
	struct User *tmp, *prev = NULL;

	for (tmp = userlist[u->hash]; tmp; tmp = tmp->next)
	{
		if (tmp == u)
		{
			if (prev)
				prev->next = tmp->next;
			else
			{
				userlist[u->hash] = tmp->next;
			}
			tmp->next = NULL;
			return;
		}
		prev = tmp;
	}
}

static struct User *new_user(char *nick)
{
	struct User *u;

	u = smalloc(sizeof(struct User));
	if (!nick)
		nick = "";
	strlcpy(u->nick, nick, MAXNICK + 1);
	add_user_to_hash_table(nick, u);

	return u;
}

#ifdef IRCD_2_10
void AddUser(char *nick, char *user, char *host, struct Server *server)
#else
void AddUser(char *nick, char *user, char *host, char *server)
#endif
{
	struct User *u;

#ifdef IRCD_2_10
	debug("AddUser(): %s (%s@%s) -> %s", nick, user, host,
			server->name);
#else
	debug("AddUser(): %s (%s@%s) -> %s", nick, user, host, server);
#endif
	u = finduser(nick);
	if (u)
	{
		log("trying to add a user that already exists? (%s)", nick);
		return;
	}

	u = new_user(nick);
	strlcpy(u->hostname, host, MAXHOST);
	strlcpy(u->username, user, MAXUSER);
#ifdef IRCD_2_10
	u->server = server;
#else
	u->server = findserver(server);
#endif
	u->t_flood = time(NULL);
	u->tld = AddTLD(u);

	if (!u->server)
		log("Unable to find server for user %s (%s?)", nick, server);
	else
		IncreaseUsers(u->server);
}

void DelUser(char *nick)
{
	struct User *u = finduser(nick);

	debug("DelUser(%s)", nick);

	if (!u)
	{
		log("DelUser(%s) failed!", nick);
		return;
	}

	del_user_from_hash_table(u);

	DecreaseUsers(u->server);

	if (u->tld)
		u->tld->users--;
	if (IsAway(u))
	{
		netinfo.away--;
#if defined(IRCD_HYB7) || defined(IRCD_2_10) || defined(IRCD_BAHAMUT) || \
    defined(IRCD_HYBSVS)
		u->server->away--;
#endif
	}
	if (IsOper(u))
	{
		netinfo.opers--;
		u->server->opers--;
		DelOperFromList(&global_oper_list, u);
	}
	if (u->myuser)
		u->myuser->flags &= ~MFLAG_ISON;

	/* Debug asserts */
	assert(u->tld->users >= 0);
	assert(u->server->opers >= 0);
	assert(netinfo.opers >= 0);
	assert(netinfo.users >= 0);

	free(u);
}

#if defined(IRCD_BAHAMUT) || defined(IRCD_HYBSVS)
void Change_User(struct User *u, char *newnick, int issvs)
#else
void Change_User(struct User *u, char *newnick)
#endif
{
	debug("Change_User(%s, %s)", u->nick, newnick);

#if defined(IRCD_BAHAMUT) || defined(IRCD_HYBSVS)
	spy("%s Change by %s to %s", issvs ? "SVSNICK" : "NICK", u->nick,
		newnick);
	if (issvs)
		sendto_dcc_flag(CF_SERVICES, "SVSNICK Change by %s to %s",
			u->nick, newnick);
	else
		sendto_dcc_flag(CF_USER, "NICK Change by %s to %s",
			u->nick, newnick);
#else
	spy("NICK Change by %s to %s", u->nick, newnick);
	sendto_dcc_flag(CF_USER, "NICK Change by %s to %s", u->nick, newnick);
#endif

	del_user_from_hash_table(u);

	strlcpy(u->nick, newnick, MAXNICK + 1);

	add_user_to_hash_table(u->nick, u);
}

#ifdef SUPPORT_HOSTSERV
void Change_Host(struct User *u, char *newhost)
{
       debug("Change_Host(%s, %s)", u->hostname, newhost);

       spy("SVSHOST Change by %s to %s", u->hostname, newhost);
        sendto_dcc_flag(CF_SERVICES, "SVSHOST Change By %s to %s",
               u->hostname, newhost);

       strlcpy(u->hostname, newhost, MAXHOST);
}

void Change_Ident(struct User *u, char *newident)
{
        debug("Change_Host(%s, %s)", u->hostname, newident);

        spy("SVSIDENt Change by %s to %s", u->username, newident);
        sendto_dcc_flag(CF_SERVICES, "SVSIDENT Change By %s to %s",
                u->username, newident);
                         
        strlcpy(u->username, newident, MAXUSER);
}
#endif

struct User *finduser(char *nick)
{
	struct User *u;

	debug("finduser(%s)", nick);

	u = userlist[HASH((unsigned char *) nick, U_TABLE_SIZE)];
	while (u && (strcasecmp(u->nick, nick) != 0))
		u = u->next;

	debug("finduser(%s) -> %s", nick, (u) ? u->nick : "NOTFOUND");

	return u;
}

void init_user_hash()
{
	int i;
	struct User *u, *prev;

	for (i = 0; i < U_TABLE_SIZE; i++)
	{
		u = userlist[i];
		while (u != NULL)
		{
			prev = u->next;
			debug("init_user_hash() deleting %s", u->nick);
			free(u);

			u = prev;
		}
		userlist[i] = NULL;
	}
	memset((void *)userlist, '\0', sizeof(userlist));
}

void UserMode(char *nick, char *modes)
{
	struct User *u = finduser(nick);
	int add = 0;

	if (!u)
	{
		log("Changing modes for unknown user: %s", nick);
		return;
	}

	while (*modes != '\0')
	{
		switch(*modes)
		{
			case '+': add = 1;	break;
			case '-': add = 0;	break;
			case 'o':
				if (add)
				{
				    if (!IsOper(u))
				    {
					u->flags |= UFLAG_ISOPER;
					AddOperToList(&global_oper_list, u);
					IncreaseOpers(u->server);
#ifdef OPER_MODES
					u->operflags |= OPER_GLOBAL;
#endif
					alert("New Oper %s on %s", u->nick,
						u->server->name);
					sendto_dcc_flag(CF_OPER,
						"New Oper %s on %s", u->nick,
						u->server->name);
				        debug("%s is now an IRCop.", u->nick);
				    }
				}
				else
				{
				    if (IsOper(u))
				    {
					u->flags &= ~UFLAG_ISOPER;
					DelOperFromList(&global_oper_list, u);
					netinfo.opers--;
					u->server->opers--;
#ifdef OPER_MODES
					u->operflags = 0;
#endif
					alert("Deoper by %s on %s", u->nick,
						u->server->name);
					sendto_dcc_flag(CF_OPER,
						"Deoper by %s on %s", u->nick,
						u->server->name);
					debug("%s is no longer an IRCop.",
						u->nick);
					assert(netinfo.opers >= 0);
					assert(u->server->opers >= 0);
				    }
				}
				break;
#ifdef IRCD_2_10
			/* IRCnet propegates AWAY by setting usermode +a */
			case 'a':
				if (add)
				{
				    if (!IsAway(u))
				    {
					u->flags |= UFLAG_ISAWAY;
					IncreaseAway(u->server);
				    }
				}
				else
				{
				    if (IsAway(u))
				    {
					u->flags &= ~UFLAG_ISAWAY;
					netinfo.away--;
					u->server->away--;
				    }
				    assert(netinfo.away >= 0);
				}
				break;
#endif
#ifdef OPER_MODES
#ifdef IRCD_BAHAMUT
			case 'A':
#else
			case 'a':
#endif
				if (add)
					u->operflags |= OPER_ADMIN;
				else
					u->operflags &= ~OPER_ADMIN;
				if (add)
				{
					alert("New Server Admin %s on %s",
						u->nick, u->server->name);
					sendto_dcc_flag(CF_OPER,
						"New Server Admin %s on %s",
						u->nick, u->server->name);
				}
				else
				{
					alert("Server Admin Status Removal %s "
						"on %s", u->nick,
						u->server->name);
					sendto_dcc_flag(CF_OPER,
						"New Server Admin %s on %s",
						u->nick, u->server->name);
				}
				break;
#ifdef IRCD_BAHAMUT
			case 'a':
				if (add)
					u->operflags |= OPER_SVSADMIN;
				else
					u->operflags &= ~OPER_SVSADMIN;
				if (add)
				{
					alert("New Services Admin %s on %s",
						u->nick, u->server->name);
					sendto_dcc_flag(CF_OPER,
						"New Services Admin %s on %s",
						u->nick, u->server->name);
				}
				else
				{
					alert("Services Admin Status Removal "
						"%s on %s", u->nick,
						u->server->name);
					sendto_dcc_flag(CF_OPER,
						"Services Admin Status Removal "
						"%s on %s", u->nick,
						u->server->name);
				}
				break;
#endif
			case 'O':
				if (add)
					u->operflags |= OPER_LOCAL;
                                else
                                        u->operflags &= ~OPER_LOCAL;
                                break;
#endif /* OPER_MODES */
			default:
				break;
		}
		modes++;
	}
}

void IncreaseUsers(struct Server *s)
{
        time_t current = time(NULL);

        debug("Increasing UserCount for %s", s->name);
        s->users++;
        netinfo.users++;
        if (s->users > s->stats->maxusers)
        {
                s->stats->maxusers = s->users;
                s->stats->t_maxusers = current;
        }
        if (netinfo.users > netinfo.maxusers)
        {
                netinfo.maxusers = netinfo.users;
                netinfo.t_maxusers = current;
        }
        if (netinfo.users > daily.users)
        {
                daily.users = netinfo.users;
                daily.t_users = current;
        }
}

