/* Stats - IRC Statistical Services for Hybrid, csircd, and IRCnet 2.10
** util.c - Misc. utility functions
**
** Copyright W. Campbell and others.  See README for more details
** Some code Copyright: Jonathan George, Kai Seidler, ircd-hybrid Team,
**                      IRCnet IRCD developers.
**
** $Id: util.c,v 1.101 2002/10/30 17:00:43 wcampbel Exp $
*/

#include "stats.h"
#include "struct.h"
#include "proto.h"
#include <signal.h>

#ifdef SPLITSERV
FILE *split_log = NULL;
#endif
FILE *main_log = NULL;

#ifdef IO_LOG
FILE *io_log = NULL;
#endif

/* print a period of seconds into human readable format */
char *timediff(time_t u)
{
        static char buf[TIMEBUF];

        if (u > 86400)
	{
		time_t days = u/86400;

                snprintf(buf, TIMEBUF, "%ld day%s, %02ld:%02ld",
                        days, (days == 1) ? "" : "s",
                        (u/3600) % 24, (u/60) % 60);
	}
        else if (u > 3600)
	{
		time_t hours = u/3600;
		time_t minutes = (u/60) % 60;

                snprintf(buf, TIMEBUF, "%ld hour%s, %ld minute%s",
                        hours, (hours == 1) ? "" : "s", minutes,
                        (minutes == 1) ? "" : "s");
	}
        else
	{
		time_t minutes = u/60;
		time_t seconds = u%60;

                snprintf(buf, TIMEBUF,
                        "%ld minute%s, %ld second%s",
                        minutes, (minutes == 1) ? "" : "s",
                        seconds, (seconds == 1) ? "" : "s");
	}

        return buf;
}

/* Generic logging routine */
void log(char *fmt, ...)
{
	va_list ap;
	char buf[BUFSIZE], fmtime[LINE];
	time_t tmp = time(NULL);

	va_start(ap, fmt);
	vsnprintf(buf, BUFSIZE, fmt, ap);

	strftime(fmtime, LINE, TIMEFMT, localtime(&tmp));

	if (main_log == NULL)
	{
		main_log = fopen(LOGFILE, "a");
		if (main_log == NULL)
			return;
	}

	fprintf(main_log, "(%s) %s\n", fmtime, buf);

	if (IsDebug())
	{
		fprintf(stderr, "(%s) %s\n", fmtime, buf);
	}

	va_end(ap);
}

#ifdef IO_LOG
void iolog(char *fmt, ...)
{
        va_list ap;
        char buf[BUFSIZE];

        va_start(ap, fmt);
        vsnprintf(buf, BUFSIZE, fmt, ap);

        if (io_log == NULL)
        {
                io_log = fopen(IOLOGFILE, "a");
                if (io_log == NULL)
                        return;
        }

        fprintf(io_log, "%s\n", buf);

        va_end(ap);
}
#endif

#ifdef SPLITSERV
/* SplitServ logging routine */
void splog(char *fmt, ...)
{
        va_list ap;
        char buf[BUFSIZE], fmtime[LINE];
        time_t tmp = time(NULL);

        va_start(ap, fmt);
        vsnprintf(buf, BUFSIZE, fmt, ap);

        strftime(fmtime, LINE, TIMEFMT, localtime(&tmp));

	if (split_log == NULL)
	{
		split_log = fopen(SPLITLOG, "a");
		if (split_log == NULL)
			return;
	}

        fprintf(split_log, "(%s) %s\n", fmtime, buf);

        if (IsDebug())
	{
                fprintf(stderr, "(%s) %s\n", fmtime, buf);
        }

        va_end(ap);
}
#endif

/* Rotate the logfile, called from chk() */
void ResetLogs()
{
	char tmp[25];
	time_t t = time(NULL);

	/* Fix the date stamp so it's the logs of THAT day, not the previous
	** day with today's date stamp on it
	*/
	t = t - 86400;
	/* It's MUCH easier to manipulate this than all the vars inside the
	** struct
	*/

	strftime(tmp, 25, BACKUPLOGFMT, localtime(&t));
	if (main_log != NULL)
	{
		fclose(main_log);
		main_log = NULL;
	}
	rename(LOGFILE, tmp);
	/* log will reopen the log file if it's closed */
	log("Started fresh logfile.");

#ifdef SPLITSERV
	strftime(tmp, 25, BACKUPSPLOGFMT, localtime(&t));
	if (split_log != NULL)
	{
		fclose(split_log);
		split_log = NULL;
	}
	rename(SPLITLOG, tmp);
	splog("Started fresh logfile.");
#endif
}

/* Our own ctime function */
/* Modified for l10n - 12 fv 2002 -Hwy */
char *sctime(time_t stuff)
{
	static char buf[32];
	struct tm *ltm = localtime(&stuff);

	/* %c is the POSIX localized replacement for ctime() */

	/* gcc complains about some non-BSD implementations not using
	** a 4 digit year in some locales.  The warning can be ignored.
	**
	** I can't find a way to shut off that gcc warning -Hwy
	*/
	strftime(buf, 32, "%c", ltm);
	return buf;
}

/* custom ftime format */
char *sftime(time_t stuff)
{
	static char fmtime[LINE];
	struct tm *ltm = localtime(&stuff);

	if (stuff != 0)
	{
		strftime(fmtime, LINE, FTMFMT, ltm);
	}
	else
	{
		snprintf(fmtime, LINE, "No Time/Record");
	}

	return fmtime;
}

char *irc_time(time_t ts)
{
	static char fmtime[LINE];

	struct tm *ltm = localtime(&ts);

	/* Since we already depend on POSIX local support, %z in the time
	** format string is acceptable...GMToffset was a nice function
	** though :)
	*/
	strftime(fmtime, LINE, IRCTFMT, ltm);

	return fmtime;
}

#if defined(CRYPT) && defined(MD5)
/* Create a random MD5 based salt for password
** encryption
*/
char* make_salt()
{
	static char saltChars[] = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789./";
	static char salt[13];
	int i;

	salt[0] = '$';
	salt[1] = '1';
	salt[2] = '$';
	for (i=3; i<=10; i++)
		salt[i] = saltChars[random() % 64];
	salt[11] = '$';
	salt[12] = '\0';

	return salt;
}
#endif

#if defined(CRYPT) && !defined(MD5)
/* create a random salt for password encryption 
** This only creates DES based salts...
**
** NOTE:  This code is partially derived from mkpasswd.c
**        by Nelson Minar (minar@reed.edu)
*/
char* make_salt()
{
	static char saltChars[] = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789./";
	static char salt[3];

	salt[0] = saltChars[random() % 64];
	salt[1] = saltChars[random() % 64];
	salt[2] = 0;

	return salt;
}
#endif /* CRYPT */

/* tokenize splits apart a message (in our case with the origin and command
** picked off already...
*/
int tokenize(char *message, char **parv)
{
        char *pos = NULL;
        char *next;
        int count = 0;

        if (!message)
        {
                /* Something is seriously wrong...bail */
                return -1;
        }

        /* First we find out if there's a : in the message, save that string
        ** somewhere so we can set it to the last param in parv
	** Also make sure there's a space before it...if not, then we're
	** screwed
        */
	pos = message;
	while (TRUE)
	{
        	if ((pos = strchr(pos, ':')))
        	{
			pos--;
			if (*pos != ' ')
			{
				/* There SHOULDN'T be a problem with this... */
				pos += 2;
				continue;
			}
                	*pos = '\0';
                	/* This is BAD - get rid of the space before the : */
                	pos++;
                	*pos = '\0';
                	/* VERY BAD...over now though */
                	pos++;
			break;
        	}
		else
		{
			break;
		}
	}

        /* Now we take the beginning of message and find all the spaces...
        ** set them to \0 and use 'next' to go through the string
        */

	next = message;
	parv[0] = message;
	count = 1;

	while (*next)
	{
		if (count == MAXPARA - 1)
		{
			/* We've reached one less than our max limit
			** to handle the parameter we already ripped off
			*/
			debug("Reached the MAXPARA limit!");
			return count;
		}
		if (*next == ' ')
		{
			*next = '\0';
			next++;
                        /* Eat any additional spaces */
                        while (*next == ' ')
                                next++;
			/* If it's the end of the string, it's simply
			** an extra space before the :parameter.  Here we
			** break.
			*/
			if (*next == '\0')
				break;
                        parv[count] = next;
                        count++;
                }
                else
                {
                        next++;
                }
        }

        if (pos)
        {
                parv[count] = pos;
                count++;
        }

        return count;
}

/* A generic tokenizer, tokenizes based on a specified character */
int generictoken(char delim, char *message, char **parv)
{
        char *next;
        int count;

        if (!message)
        {
                /* Something is seriously wrong...bail */
                return -1;
        }

        /* Now we take the beginning of message and find all the spaces...
        ** set them to \0 and use 'next' to go through the string
        */
	next = message;
	parv[0] = message;
	count = 1;

        while (*next)
        {
		/* This is fine here, since we don't have a :delimited
		** parameter like tokenize
		*/
        	if (count == MAXPARA)
		{
		    /* We've reached our limit */
		    debug("Reached the MAXPARA limit!");
		    return count;
		}
		if (*next == delim)
		{
			*next = '\0';
			next++;
			/* Eat any additional delimiters */
			while (*next == delim)
				next++;
                        /* If it's the end of the string, it's simply
                        ** an extra space at the end.  Here we break.
                        */
                        if (*next == '\0')
                                break;
			parv[count] = next;
			count++;
		}
		else
		{
			next++;
		}
        }

	return count;
}

/* our own malloc with logging, sanity checking, and zeroing out memory */
void *smalloc(size_t size)
{
        void *buf;

        if (!size)
	{
                log("smalloc(): illegal attempt to allocate 0 bytes!");
                size = 1;
        }
        buf = calloc(1, size);
        if (!buf)
	{
                log("smalloc(): out of memory.");
                exit(0);
        }
        return buf;
}

/* Generate a hash value into the tables */
unsigned long HASH(const unsigned char *name, int size_of_table)
{
        unsigned long h = 0, g;

        while (*name)
	{
                h = (h << 4) + *name++;
                        if ((g = (h & 0xF0000000)))
                                h ^= g >> 24;
                        h &= ~g;
        }
        return h % size_of_table;
}

/* Lowercase a string */
char *strlower(char *s)
{
        char *t = s;
        while ((*t++ = tolower(*t)))
                ;
        return s;
}

/* Uppercase a string */
char *strupper(char *s)
{
	char *t = s;
	while ((*t++ = toupper(*t)))
		;
	return s;
}

/* Get the version flags string */
char *VersionFlags()
{
	static char ver[LINE];
	char *verptr = ver;

#ifdef TRACK_ADMIN
	*verptr++ = 'A';
#endif
#ifdef SUPPORT_MSG_UH
	*verptr++ = 'U';
#endif
#ifdef SUPPORT_MSG_OPERS
	*verptr++ = 'O';
#endif
#ifdef IGNORE_ILLEGAL_TLD
	*verptr++ = 'I';
#endif
#ifdef TRACK_VERSION
	*verptr++ = 'V';
#endif
#ifdef JUPESERV
	*verptr++ = 'J';
#endif
#ifdef ENABLE_QUOTE
	*verptr++ = 'Q';
#endif
#ifdef DELAY_PRIVMSG
	*verptr++ = 'D';
#endif
#ifdef LARGE_NET
	*verptr++ = 'N';
#endif
#ifdef HIDE_IPS
        *verptr++ = 'H';
#endif
#ifdef SUPPORT_HOSTSERV
	*verptr++ = 'h';
#endif
#ifdef OPER_MODES
	*verptr++ = 'M';
#endif
#ifdef DIE_ON_MASS_COLLIDE
	*verptr++ = 'C';
#endif
#ifdef DISALLOW_NICK_PRIVMSG
	*verptr++ = 'u';
#endif
#ifdef SPLITSERV
	*verptr++ = 'P';
#endif
#ifdef STATSERV
	*verptr++ = 'T';
#endif
#ifdef ALERTMODE
	*verptr++ = 'a';
#endif
#ifdef SPYMODE
	*verptr++ = 's';
#endif

	if (IsDebug())
		*verptr++ = 'd';
	if (IsNoFork())
		*verptr++ = 'n';

	*verptr++ = ' ';
	*verptr = '\0';

#ifdef IRCD_HYBRID
	strlcat(ver, "HYBRID", LINE);
#endif
#ifdef IRCD_HYB6
	strlcat(ver, "HYB6", LINE);
#endif
#ifdef IRCD_HYB7
	strlcat(ver, "HYB7", LINE);
#endif
#ifdef IRCD_2_10
	strlcat(ver, "2_10", LINE);
#endif
#ifdef IRCD_BAHAMUT
	strlcat(ver, "BAHAMUT", LINE);
#endif
#ifdef IRCD_HYBSVS
	strlcat(ver, "HYBSVS", LINE);
#endif

	return ver;
}

#ifdef OPER_MODES
char opermode(struct User *u)
{
	if (u == NULL)
		return '-';
	if (u->operflags & OPER_MYOPER)
		return '*';
#ifdef IRCD_BAHAMUT
	if (u->operflags & OPER_SVSADMIN)
		return 'a';
#endif
	if (u->operflags & OPER_ADMIN)
		return 'A';
	if (u->operflags & OPER_GLOBAL)
		return 'O';
	if (u->operflags & OPER_LOCAL)
		return 'o';
	/* Shouldn't get here */
	return 'x';
}
#endif /* OPER_MODES */

void SigHup(int sig)
{
	need_rehash = 1;
	signal(SIGHUP, SigHup);
}

void SigUsr1(int sig)
{
	/* This should be sufficient to tell it to reconnect */
	close(servsock);
	delete_connection(servsock);
	servsock = -1;
	signal(SIGUSR1, SigUsr1);
}

void SigTerm(int sig)
{
        wallops(settings.name, "Received SIGTERM: Exiting");
	alert("Received SIGTERM: Exiting");
	sendto_dcc_flag(CF_NOTICES, "Received SIGTERM: Exiting");
#ifdef STATSERV
        sts(":%s QUIT :Received SIGTERM: Exiting", StatServ.nick);
#endif
	sts(":%s QUIT :Received SIGTERM: Exiting", AdminServ.nick);
#ifdef SPLITSERV
	sts(":%s QUIT :Received SIGTERM: Exiting", SplitServ.nick);
#endif
#ifdef JUPESERV
	sts(":%s QUIT :Received SIGTERM: Exiting", JupeServ.nick);
#endif
        sleep(1);
        close(servsock);
	close_dccs();
        SaveStats();
        log("Received SIGTERM: Exiting");
	/* No need to reinstall the signal handler */
        exit(0);
}

int is_channel(char *chan)
{
        if (chan != NULL)
	{
		/* Sadly, a few servers propogate &channels... CSr comes to
		** mind here.
		*/
                if ( (chan[0] == '#') || (chan[0] == '+') || (chan[0] == '&') )
                        return TRUE;
                return FALSE;
        }

        /* Shouldn't get here */
        return FALSE;
}

void alert(char *fmt, ...)
{
#ifdef ALERTMODE
	va_list ap;
	char buf[BUFSIZE];

	if (ad_delayed)
		return;
	va_start(ap, fmt);
	vsnprintf(buf, sizeof(buf), fmt, ap);
	/* snprintf will always NULL terminate */
	privmsg(ALERTCHAN, AdminServ.nick, MSG_PRIVMSG, "%s", buf);
	va_end(ap);
#endif
}

void spy(char *fmt, ...)
{
#ifdef SPYMODE
        va_list ap;
        char buf[BUFSIZE];

	if (ad_delayed)
		return;
        va_start(ap, fmt);
        vsnprintf(buf, sizeof(buf), fmt, ap);
        privmsg(SPYCHAN, AdminServ.nick, MSG_PRIVMSG, "%s", buf);
        va_end(ap);
#endif
}

void debug(char *fmt, ...)
{
	va_list ap;
	char buf[BUFSIZE];

	if (IsDebug())
	{
		va_start(ap, fmt);
		vsnprintf(buf, sizeof(buf), fmt, ap);
		log("DEBUG: %s", buf);
		va_end(ap);
	}
}

void InitLists(void)
{
	hiddenhead.head = NULL;
	hiddenhead.tail = NULL;
#ifdef IRCD_2_10
        servicehead.head = NULL;
	servicehead.tail = NULL;
#endif
	global_oper_list.head = NULL;
	global_oper_list.tail = NULL;
}

void ExpandParv(char *dest, int max, int initial, int parc, char *parv[])
{
	int i;

	if (parv[initial])
		strlcpy(dest, parv[initial], max);
	for (i = initial + 1; i < parc; i++)
	{
		strlcat(dest, " ", max);
		strlcat(dest, parv[i], max);
	}
}

/* Is it the midnight hour */
int is_midnight()
{
        time_t current = time(NULL);
        struct tm *ltm = localtime(&current);

        if (ltm->tm_hour == 0)
                return 1;

        return 0;
}

#if defined(SPLITSERV) || defined(DELAY_PRIVMSG)
/* Perform a 3/4 of a second delay using select() */
void select_delay(void)
{
	struct timeval tv;

	tv.tv_usec = 750000; /* 3/4 a second delay */
	tv.tv_sec = 0;       /* ignore this */
	select(1, NULL, NULL, NULL, &tv);
}
#endif
